'''
Module to test the post_process module.
'''

import pytest

from rheia.POST_PROCESS.post_process import PostProcessOpt, PostProcessUQ


@pytest.fixture
def input_case_uq():
    """
    Instantiate a PostProcessUQ object.

    Returns
    -------
    test_obj : object
        PostProcessUQ object.

    """
    case = 'H2_FUEL'
    test_obj = PostProcessUQ(case, 2)

    return test_obj


def test_get_fitness_population():
    """
    Assert the population and fitness samples imported from fitness and
    population files.
    """

    case = 'H2_FUEL'
    eval_type = 'DET'
    test_obj = PostProcessOpt(case, eval_type)

    x, y = test_obj.get_fitness_population('run_tutorial')

    assert round(x[0][0], 3) == 7.796
    assert round(x[0][12], 3) == 11.117
    assert round(x[1][7], 3) == 106.956
    assert round(y[0][0], 3) == 1.933
    assert round(y[0][12], 3) == 4.382
    assert round(y[1][7], 3) == 2.959


def test_get_pdf(input_case_uq):
    """
    Assert the data points for the Probability Density Function, stored in a
    UQ results folder.

    Parameters
    ----------
    input_case_uq : object
        PostProcessUQ object.

    """

    result_dir = 'opt_design_tutorial'
    objective = 'lcoh'
    x, y = input_case_uq.get_pdf(result_dir, objective)

    assert round(x[0], 3) == 5.187
    assert round(x[5], 3) == 5.481
    assert round(x[7], 3) == 5.598
    assert round(y[23], 3) == 0.205
    assert round(y[29], 3) == 0.324
    assert round(y[35], 3) == 0.415


def test_get_cdf(input_case_uq):
    """
    Assert the data points for the Cumulative Distribution Function, stored in
    a UQ results folder.

    Parameters
    ----------
    input_case_uq : object
        PostProcessUQ object.

    """

    result_dir = 'opt_design_tutorial'
    objective = 'lcoh'
    x, y = input_case_uq.get_cdf(result_dir, objective)

    assert round(x[0], 3) == 5.187
    assert round(x[5], 3) == 5.481
    assert round(x[7], 3) == 5.598
    assert round(y[23], 3) == 0.084
    assert round(y[29], 3) == 0.180
    assert round(y[35], 3) == 0.315


def test_get_loo(input_case_uq):
    """
    Assert the Leave-One-Out error, stored in a UQ results folder.

    Parameters
    ----------
    input_case_uq : object
        PostProcessUQ object.

    """

    result_dir = 'opt_design_tutorial'
    objective = 'lcoh'
    loo = input_case_uq.get_loo(result_dir, objective)

    assert round(loo, 6) == 0.005952


def test_get_sobol(input_case_uq):
    """
    Assert the Sobol' indices, stored in a UQ results folder.

    Parameters
    ----------
    input_case_uq : object
        PostProcessUQ object.

    """

    result_dir = 'opt_design_tutorial'
    objective = 'lcoh'
    names, sobol = input_case_uq.get_sobol(result_dir, objective)

    assert names[0] == 'capex_pemel'
    assert names[5] == 'opex_pemel'
    assert round(sobol[0], 3) == 0.327
    assert round(sobol[5], 3) == 0.057



'''
Module to test the det_stoch_des_space module.
'''

import pytest
import numpy as np
import rheia.OPT.optimization as opt


@pytest.fixture
def run_dict():
    """
    Generate a generic input dictionary for optimization.

    Returns
    -------
    dict_opt : dict
        Input dictionary for optimization.

    """

    dict_opt = {'case': 'H2_FUEL',
                'objectives': {'DET': (-1, 1)},
                'stop': ('BUDGET', 2000),
                'population size': 20,
                'results dir': 'run_tutorial',
                }

    return dict_opt


@pytest.fixture
def input_case(run_dict):
    """
    Instantiate a StochasticDesignSpace object, get the function
    to evaluate the model and the list of fixed parameters.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.

    Returns
    -------
    space_obj : object
        StochasticDesignSpace object.
    params : list
        List of fixed parameters for evaluation of the model.
    eval_func : function
        Function for evaluation of the model.

    """

    design_space = 'design_space.csv'
    space_obj, eval_func, params = opt.load_case(run_dict, design_space)

    return space_obj, eval_func, params


def test_parse_available_opt():
    """
    Assert if the NSGA2 method is available in the package.
    """

    methods = opt.parse_available_opt()
    assert methods[0][0] == 'NSGA2'


def test_load_optimizer():
    """
    Assert the name for a test object of the NSGA2 class.
    """

    opt_obj = opt.load_optimizer('NSGA2')
    assert opt_obj.__name__ == 'NSGA2'


def test_check_existing_results(run_dict, input_case):
    """
    Assert function that checks if existing results are present.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.
    input_case : list
        List that includes the StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    bool_test = opt.check_existing_results(run_dict, input_case[0])
    assert bool_test


def test_check_existing_results_false(run_dict, input_case):
    """
    Assert function that checks if existing results are present.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.
    input_case : list
        List that includes the StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    run_dict['results dir'] = 'run_test_no_results'
    bool_test = opt.check_existing_results(run_dict, input_case[0])
    assert bool_test is False


def test_scale_samples_to_design_space(input_case):
    """
    Assert function that scales samples to the design space.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.
    input_case : list
        List that includes the StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    nondim_doe = np.array([[0.5, 0.5], [0.1, 0.9]])
    dim_doe = opt.scale_samples_to_design_space(nondim_doe, input_case[0])
    assert round(dim_doe[0][0], 1) == 3.5
    assert round(dim_doe[0][1], 1) == 3.5
    assert round(dim_doe[1][0], 1) == 0.7
    assert round(dim_doe[1][1], 1) == 6.3



"""
The :py:mod:`h2_power` module contains a class to read the required data and
a class to evaluate the power-to-power system.
"""

import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pvlib


class ReadData:
    """

    This class enables to read data from the data files.

    Parameters
    ----------
    filename_climate : str
        The directory of the file with information on the
        solar irradiance.
    filename_climate : str
        The directory of the file with information on the
        demand.

    """

    def __init__(self, filename_climate, filename_demand):
        self.filename_climate = filename_climate
        self.filename_demand = filename_demand
        self.path = os.path.dirname(os.path.abspath(__file__))

    def load_climate(self):
        """

        This method loads the hourly solar irradiance data
        and ambient temperature data,
        situated in the 'sol_irr' and 'T_amb' columns of the
        climate data file.

        Returns
        -------
        sol_irr : ndarray
            The hourly solar irradiance data for a Typical
            Meteorological Year. (8760 elements)
        t_amb : ndarray
            The hourly ambient temperature data for a Typical
            Meteorological Year. (8760 elements)

        """
        data = pd.read_csv(self.filename_climate)
        sol_irr = data['sol_irr'].to_numpy()
        t_amb = data['T_amb'].to_numpy()

        return sol_irr, t_amb

    def load_demand(self):
        """

        This method loads the hourly electricity demand data,
        situated in the 'total electricity' column of the demand data file.

        Returns
        -------
        load_elec : ndarray
            The hourly electricity demand.

        """
        data = pd.read_csv(self.filename_demand)
        load_elec = data['total electricity'].to_numpy() * 1e3

        return load_elec

    def load_parameters(self):
        """

        This method loads the deterministic values of the model
        parameters, defined in the design_space file. This is
        useful when the deterministic performance of a specific
        design needs to be evaluated.

        Returns
        -------
        param_dict : dict
            Dictionary with the names of the model parameters
            and the corresponding deterministic values.

        """
        param_dict = {}
        design_space = os.path.join(self.path, 'design_space.csv')

        # read the deterministic values for the parameters in `design_space`
        with open(design_space, 'r') as file:
            for line in file:
                tmp = line.split(",")
                if tmp[1] == 'par':
                    param_dict[tmp[0]] = float(tmp[2])

        return param_dict


class Evaluation:
    """

    This class evaluates the photovoltaic-hydrogen system.
    For a given design, the solar irradiance, electricity demand
    and the characterization of the model parameters,
    the levelized cost of electricity and the self-sufficiency ratio
    are quantified.

    Parameters
    ----------
    sol_irr : ndarray
        The hourly solar irradiance for the evaluated year.
    t_amb : ndarray
        The hourly ambient temperature for the evaluated year.
    L_elec : ndarray
        The hourly electricity demand for the evaluated year.
    parameters : dict
        Dictionary with the model parameters and design variables values.

    """

    def __init__(self, sol_irr, t_amb, L_elec, par):
        self.par = par

        # the solar irradiance and ambient temperature are scaled with the
        # corresponding uncertainty
        self.sol_irr = sol_irr * self.par['u_sol_irr']
        self.t_amb = t_amb + self.par['u_t_amb']

        # the electric load, scaled with the uncertainty on the load
        self.load_elec = L_elec * self.par['u_load_elec']
        # the result dictionary
        self.res = {}

        # the system lifetime
        self.par['life_sys'] = 20.

        # initialize the operating hours of the electrolyzer and fuel cell
        self.res['running_hours_pemel'] = 0.
        self.res['running_hours_pemfc'] = 0.

        # initialize the storage tank size and its starting status
        self.m_h2_max = self.tank()
        self.m_h2_min = 0.05 * self.m_h2_max
        self.m_h2 = self.m_h2_min

        # the number of PEM electrolyzer cells and PEM fuel cells,
        # corresponding to the nominal capacity of the considered
        # PEM electrolyzer and PEM fuel cell and the provided
        # system capacities
        self.n_pemel_array = self.par['n_pemel'] / 0.4
        self.n_pemfc_array = self.par['n_pemfc'] / (5. / 35.)

        # generate the fitted polynomial on the electrolyzer array and
        # fuel cell array power - current relation
        self.polyfit_pemel()
        self.polyfit_pemfc()

        # instantiate the profiles for grid electricity price
        self.elec_profiles()

    def elec_profiles(self):
        """
        Set the grid electricity price for buying and selling electricity.
        A contract with fixed electricity price is considered, for which the
        price for buying electricity consists of three segments: the energy
        price itself (i.e. 'elec cost'), the profit made on this price by the
        electricity provider (i.e. 'elec_cost_profit') and the fraction of the
        energy price to the final retail price (i.e. 'elec_cost_ratio', e.g.
        when this value equal 0.3, then the energy price corresponds to 30% of
        the final bill, while 70% corresponds to transportation cost,
        taxes,...). The price for selling electricity back to the grid
        corresponds to the energy price.

        """

        self.elec_profile = np.ones(len(self.sol_irr)) * (
            (self.par['elec_cost'] + self.par['elec_cost_profit']) /
            self.par['elec_cost_ratio']) / 1e6

        self.elec_profile_sale = np.ones(len(self.sol_irr)) * (
            self.par['elec_cost'] / 1e6)

    #############################
    # photovoltaic array module #
    #############################

    def quantify_mpp(self, sol_irr, t_amb, pv_system):
        """

        Quantify the maximum power of the photovoltaic array
        for a given solar irradiance and ambient temperature.

        Parameters
        ----------
        sol_irr : float
            The solar irradiance [W/m2].
        t_amb : float
            The ambient temperature [C].
        pv_system : pandas.core.series.Series
            The pv system characteristics
        Returns
        -------
        pmp : float
            The maximum power.

        """

        # quantify the parameters for the pv system using De Soto method
        pv_inputs = pvlib.pvsystem.calcparams_desoto(sol_irr,
                                                     t_amb,
                                                     pv_system['alpha_sc'],
                                                     pv_system['a_ref'],
                                                     pv_system['I_L_ref'],
                                                     pv_system['I_o_ref'],
                                                     pv_system['R_sh_ref'],
                                                     pv_system['R_s'],
                                                     EgRef=1.121,
                                                     dEgdT=-0.0002677,
                                                     irrad_ref=1000.,
                                                     temp_ref=25.)

        # determine the maximum power for the given pv system
        pmp = pvlib.pvsystem.max_power_point(pv_inputs[0],
                                             pv_inputs[1],
                                             pv_inputs[2],
                                             pv_inputs[3],
                                             pv_inputs[4],
                                             method='newton')['p_mp']

        return pmp

    def photovoltaic(self):
        """

        The hourly photovoltaic power is quantified via the PVlib package.
        Using this package, first the characteristics for a typical
        photovoltaic panel are defined. Based on these characteristics,
        the maximum power point is quantified for each hour, based on the
        corresponding solar irradiance and ambient temperature. Finally, the
        hourly power production is scaled by the considered photovoltaic array
        capacity.

        """

        p_pv = np.zeros(len(self.sol_irr))

        # get the specific photovoltaic panel characteristics
        pv_database = pvlib.pvsystem.retrieve_sam('CECmod')
        pv_system = pv_database.SunPower_SPR_X19_240_BLK

        p_mpp_ref = self.quantify_mpp(1000., 25., pv_system)  # W

        # maximum power point determination for each hour in the timeframe
        for i, irr in enumerate(self.sol_irr):
            if irr > 0.:
                p_mpp = self.quantify_mpp(irr, self.t_amb[i], pv_system)
                p_pv[i] = p_mpp / p_mpp_ref * self.par['n_pv'] * 1e3  # W
            else:
                p_pv[i] = 0.

        # store the hourly pv power in the result dictionary
        self.res['p_pv'] = p_pv

        # the dc-dc converter capacity in kW
        self.res['n_dcdc_pv'] = max(p_pv) / 1e3

    def net_power(self):
        """

        Determine the hourly net power.
        This corresponds to the net power available (or still required) on the
        DC bus bar after extracting the load from the photovoltaic power.

        Returns
        -------
        p_net : ndarray
            The hourly net power. [W]

        """

        p_net = np.zeros(len(self.sol_irr))

        # determine the hourly photovoltaic power
        self.photovoltaic()

        for i, p_pv in enumerate(self.res['p_pv']):

            # the excess pv power after extracting the electric load
            p_net[i] = p_pv - self.load_elec[i]

        return p_net

    ###############
    # tank module #
    ###############

    def tank(self):
        """

        The maximum storage capacity of the hydrogen tank.

        Returns
        -------
        m_max : float
            The hydrogen storage capacity. [kg]

        """

        # conversion from energy (kWh) into mass (kg)
        m_max = self.par['n_tank'] / 33.33

        return m_max

    #############################
    # electrolyzer array module #
    #############################

    def pemel(self, i_pemel):
        """
        The electrolyzer model, based on the work of Saeed et al. [1]. For a
        given current, the model determines the operating voltage by
        considering the activation, concentration and ohmic overpotentials.
        The model quantifies the operating voltage, power, efficiency and
        hydrogen production.

        [1] Saeed, E. W., & Warkozek, E. G. (2015). Modeling and Analysis of
            Renewable PEM Fuel Cell System. Energy Procedia, 74, 87–101.
            https://doi.org/10.1016/j.egypro.2015.07.527

        Parameters
        ----------
        i_pemel : float
            The electrolyzer input current [A].

        Returns
        -------
        res : dict
            Dictionary with the operating conditions of the electrolyzer for a
            given current. It contains items on the operating voltage, power,
            efficiency and hydrogen mass flow rate.

        """
        par_pemel = {'T': 353.,
                     'a': 1.,
                     'p_o2': 1.,
                     'p_h2': 1.,
                     'p_h2o': 1.,
                     'i_L': 2.,
                     'A': 100.,
                     'i_0': 1e-4,
                     'n': 2.,
                     't_mem': 50e-4,
                     'alpha': 0.3,
                     'R': 8.3143,
                     'F': 96485.,
                     'HHV': 141.7e6,
                     }

        res = {}
        i = i_pemel / par_pemel['A']

        # minimum operating voltage of electrolyzer
        e_0 = (1.48 - 0.85e-3 * (par_pemel['T'] - 298.15) + 4.3085e-5 *
               par_pemel['T'] * np.log(par_pemel['p_h2'] *
                                       np.sqrt(par_pemel['p_o2']) /
                                       par_pemel['p_h2o']))

        # activation overpotential
        v_act = (np.log(i / par_pemel['i_0']) /
                 (par_pemel['alpha'] * par_pemel['n'] * par_pemel['F']) *
                 par_pemel['R'] * par_pemel['T'])

        # ohmic overpotential
        lambda_mem = (0.043 + 17.81 * par_pemel['a'] -
                      39.85 * par_pemel['a']**2. +
                      36. * par_pemel['a']**3.)
        sigma_mem = ((0.005139 * lambda_mem - 0.00326) *
                     np.exp(1268 * (1. / 303. - 1. / par_pemel['T'])))
        v_ohm = i * par_pemel['t_mem'] / sigma_mem

        # the concentration overpotential
        v_con = - (par_pemel['R'] * par_pemel['T'] /
                   (par_pemel['n'] * par_pemel['F']) *
                   np.log(1. - i / par_pemel['i_L']))

        # model outputs
        res['v_pemel'] = (e_0 + v_act + v_ohm + v_con) * self.n_pemel_array
        res['m_pemel'] = self.current_to_mh2(i_pemel) * self.n_pemel_array
        res['p_pemel'] = i_pemel * res['v_pemel']
        res['eff_pemel'] = (res['m_pemel'] * par_pemel['HHV'] /
                            (res['p_pemel'] * 3600.))
        return res

    def current_to_mh2(self, current):
        """
        When current is provided, this function determines the
        corresponding hydrogen mass flow rate per hour.

        Parameters
        ----------
        current : float
            The electrolyzer input current [A].

        Returns
        -------
        m_h2 : float
            The produced hydrogen mass flow rate [kg/h].

        """
        far_cons = 96485.
        m_h2 = current / (2. * far_cons) * 2.02e-3 * 3600.

        return m_h2

    def polyfit_pemel(self):
        """
        The electrolyzer stack is evaluated over a range of input currents.
        Following these evaluations, a polynomial is fitted on the
        power - current relation of the electrolyzer. This polynomial enables
        to rapidly determine the input current when a certain amount of power
        is available. Since this relation is fairly linear, the polynomial
        should reach good agreement with the actual power - current relation,
        while maintaining the level of fidelity of the actual model.

        """

        # evaluate the electrolyzer stack for a set of currents
        i_list = np.arange(start=3, stop=200, step=4)
        p_pemel = np.zeros(len(i_list))
        for index, i in enumerate(i_list):
            res = self.pemel(i)
            p_pemel[index] = res['p_pemel']

        # generate a polynomial fitted on the power - current points
        self.p_to_i_pemel = polyfit_func(p_pemel, i_list)

    def charge_pemel(self, p_pemel):
        """
        For a given power supplied to the electrolyzer, this function
        determines the actual hydrogen produced. First, the method evaluates
        if the power supplied lies within the operating bounds of the
        electrolyzer stack. If the power is situated below the lower limit,
        the electrolyzer does not run. Instead, when the power is situated
        above the upper limit, the electrolyzer operates at nominal conditions.
        The input current is determined through the fitted polynomial on the
        power - current relation of the electrolyzer stack. As the power is an
        output from the electrolyzer model, considering this polynomial avoids
        the use of root finding methods and is therefore more robust in
        optimization and uncertainty quantification approaches.

        When the hydrogen mass is determined that can be generated, the current
        hydrogen tank capacity is addressed. If the capacity exceeds the
        maximum storage tank capacity, the current applied to the electrolyzer
        is re-evaluated, such that this current matches the hydrogen production
        that leads to a full hydrogen tank.

        Finally, when hydrogen is produced, the
        running hours of the electrolyzer stack is increased by 1 and the
        power consumed by the electrolyzer array is returned.

        Parameters
        ----------
        p_pemel : float
            The power available for electrolysis [W].

        Returns
        -------
        p_consumed : float
            The power consumed by the electrolyzer array [W].

        """

        # the operating bounds
        op_lower_lim = self.par['n_pemel'] * 10.
        op_upper_lim = self.par['n_pemel'] * 1e3

        # check if power is higher than the lowest operating point
        if (self.m_h2 < self.m_h2_max and
                p_pemel > op_lower_lim and
                self.par['n_pemel'] > 1e-1):

            # if the power exceeds the upper bound, operate at upper bound
            p_pemel_applied = min(p_pemel, op_upper_lim)

            # current at this applied power
            i_pemel = self.p_to_i_pemel(p_pemel_applied)
            m_h2 = self.current_to_mh2(i_pemel) * self.n_pemel_array
            p_consumed = p_pemel_applied

            # produced hydrogen is added to the tank
            self.m_h2 += m_h2

            # if the new hydrogen capacity exceeds the maximum storage capacity
            if self.m_h2 > self.m_h2_max:

                # define current that results in a full storage tank
                excess = self.m_h2 - self.m_h2_max
                addition = m_h2
                allowed_current = (1. - excess / addition) * i_pemel

                # check if this current is still higher than lower limit
                if allowed_current > self.p_to_i_pemel(op_lower_lim):
                    p_consumed = self.pemel(allowed_current)['p_pemel']
                    self.m_h2 = self.m_h2_max
                    self.res['running_hours_pemel'] += 1.
                else:
                    self.m_h2 -= m_h2
                    p_consumed = 0.
            else:
                self.res['running_hours_pemel'] += 1.
        else:
            p_consumed = 0.

        return p_consumed

    ##########################
    # fuel cell array module #
    ##########################

    def polyfit_pemfc(self):
        """
        The fuel cell stack is evaluated over a range of input currents.
        Following these evaluations, a polynomial is fitted on the
        power - current relation of the fuel cell. This polynomial enables
        to rapidly determine the input current when a certain amount of power
        needs to be provided. Since this relation is fairly linear, the
        polynomial should reach good agreement with the actual power - current
        relation, while maintaining the level of fidelity of the actual model.

        """

        # evaluate the electrolyzer stack for a set of currents
        i_list = np.arange(start=1, stop=325, step=4)
        p_pemfc = np.zeros(len(i_list))
        for index, i in enumerate(i_list):
            res = self.pemfc(i)
            p_pemfc[index] = res['p_pemfc']

        # generate a polynomial fitted on the power - current points
        self.p_to_i_pemfc = polyfit_func(p_pemfc, i_list)

    def pemfc(self, i_pemfc):
        """
        The PEM fuel cell model, based on the work of Murugesan et al. [2].
        The model determines the current-voltage characteristic and provides
        the voltage, power, efficiency and hydrogen consumption.

        [2] Murugesan, K., & Senniappan, V. (2013). Investigation of water
        management dynamics on the performance of a Ballard-Mark-V proton
        exchange membrane fuel cell stack system. International Journal of
        Electrochemical Science, 8(6), 7885–7904.

        Parameters
        ----------
        i_pemfc : float
            The fuel cell output current.

        Returns
        -------
        res : dict
            Dictionary with the operating conditions of the fuel cell for a
            given current. It contains items on the operating voltage, power,
            efficiency and hydrogen mass flow rate.

        """

        par_pemfc = {'p_o2': 3.,
                     'p_h2': 3.,
                     'i_L': 1.5,
                     'T': 353.,
                     'A': 232.,
                     't_mem': 178e-4,
                     'F': 96485.,
                     'B': 0.016,
                     'HHV': 141.7e6,
                     'lambda_mem': 23.,
                     'eps_1': -0.948,
                     'eps_2': 0.00354,
                     'eps_3': 7.6e-5,
                     'eps_4': -1.93e-4,
                     'b2': 1268.,
                     'b11': 0.005139,
                     'b12': 0.00326,
                     }

        res = {}
        i = i_pemfc / par_pemfc['A']

        # minimum operating voltage of fuel cell
        e_0 = (1.229 - 0.85e-3 * (par_pemfc['T'] - 298.15) + 4.31e-5 *
               par_pemfc['T'] * np.log(par_pemfc['p_h2'] *
                                       np.sqrt(par_pemfc['p_o2'])))

        # activation overpotential
        c_o2 = par_pemfc['p_o2'] / (5.08e6 * np.exp(498. / par_pemfc['T']))
        v_act = - (par_pemfc['eps_1'] + par_pemfc['eps_2'] * par_pemfc['T'] +
                   par_pemfc['eps_3'] * par_pemfc['T'] * np.log(c_o2) +
                   par_pemfc['eps_4'] * np.log(i_pemfc) * par_pemfc['T'])

        # ohmic overpotential
        b_1 = par_pemfc['b11'] * par_pemfc['lambda_mem'] - par_pemfc['b12']
        sigma_mem = b_1 * np.exp(par_pemfc['b2'] *
                                 (1 / 303. - 1. / par_pemfc['T']))
        r_m = par_pemfc['t_mem'] / sigma_mem
        v_ohm = i_pemfc * r_m / par_pemfc['A']

        # concentration overpotential
        v_con = - par_pemfc['B'] * np.log(1. - i / par_pemfc['i_L'])

        # model outputs
        res['v_pemfc'] = (e_0 - v_act - v_ohm - v_con) * self.n_pemfc_array
        res['m_pemfc'] = self.current_to_mh2(i_pemfc) * self.n_pemfc_array
        res['p_pemfc'] = i_pemfc * res['v_pemfc']
        res['eff_pemfc'] = (res['m_pemfc'] * par_pemfc['HHV'] /
                            (res['p_pemfc'] * 3600.))
        return res

    def charge_pemfc(self, p_pemfc):
        """

        This method evaluates if the power required from the fuel cell
        lies within the operating range of the DC-DC converter and of the
        fuel cell array. If yes, the power is supplied and the
        consumed hydrogen is quantified. If the required power is larger than
        the fuel cell capacity, than the nominal power is supplied.
        If the consumed hydrogen is larger than the available hydrogen in
        the hydrogen storage tank, the supplied power is recalculated,
        such that the consumed hydrogen matches the initial available hydrogen
        left in the hydrogen storage tank.
        Finally, the operating hours of the fuel cell array is
        increased by 1.

        Parameters
        ----------
        p_pemfc : float
            The power demanded from the fuel cell array.

        Returns
        -------
        p_produced : float
            The actual power produced by the fuel cell array.
        """

        # the operating bounds
        op_lower_lim = self.par['n_pemfc'] * 10.
        op_upper_lim = self.par['n_pemfc'] * 1e3

        # check if power is higher than the lowest operating point
        if (self.m_h2 > self.m_h2_min and
                p_pemfc > op_lower_lim and
                self.par['n_pemfc'] > 1e-1):

            # if the power exceeds the upper bound, operate at upper bound
            p_pemfc_applied = min(p_pemfc, op_upper_lim)

            # current at this required power
            i_pemfc = self.p_to_i_pemfc(p_pemfc_applied)
            m_h2 = self.current_to_mh2(i_pemfc) * self.n_pemfc_array
            p_produced = p_pemfc_applied

            # produced hydrogen is extracted from the tank
            self.m_h2 -= m_h2

            # if the new hydrogen capacity is below the minimum storage
            # capacity
            if self.m_h2 < self.m_h2_min:

                # define current that results in an empty storage tank
                lack = self.m_h2_min - self.m_h2
                extracted = m_h2
                allowed_current = (1. - lack / extracted) * i_pemfc

                # check if this current is still higher than lower limit
                if allowed_current > self.p_to_i_pemfc(op_lower_lim):
                    p_produced = self.pemfc(allowed_current)['p_pemfc']
                    self.m_h2 = self.m_h2_min
                    self.res['running_hours_pemfc'] += 1.
                else:
                    self.m_h2 += m_h2
                    p_produced = 0.
            else:
                self.res['running_hours_pemfc'] += 1.
        else:
            p_produced = 0.

        return p_produced

    #####################
    # evaluation module #
    #####################

    def evaluation(self):
        """

        This is the main method of the Evaluation class.
        For each hour, the power management strategy is applied.
        If the net power is positive, the electrolyzer is charged. If, after
        the charging, there is still power available, this power is sold to
        the grid. Instead, when the net power is negative, the fuel cell is
        charged. If the power generated by the fuel cell is insufficient, the
        remaining power is extracted from the grid.
        Finally, the self-sufficiency ratio and the system cost are determined.

        """

        n_dcdc_pemel = np.zeros(len(self.sol_irr))
        n_dcdc_pemfc = np.zeros(len(self.sol_irr))
        n_dcac = np.zeros(len(self.sol_irr))

        self.res['m_h2_array'] = np.ones(len(self.sol_irr))
        self.res['grid_e_buy'] = np.ones(len(self.sol_irr))
        self.res['grid_e_sold'] = np.ones(len(self.sol_irr))

        # the net power, i.e. after using the pv power to cover the load
        p_net = self.net_power()

        for t, p_net in enumerate(p_net):
            e_grid_buy = 0.
            e_grid_sold = 0.

            # if excess power available, charge electrolyzer
            if p_net > 0.:
                p_consumed = self.charge_pemel(p_net)
                n_dcdc_pemel[t] = p_consumed
                p_rem = p_net - p_consumed

                # sell remaining electricity to the grid
                e_grid_sold += p_rem

                # power that goes through the DC-AC converter,
                # i.e. the load is covered by the pv array +
                # additional energy is sold to the grid
                n_dcac[t] = self.load_elec[t] + e_grid_sold

            # if load remaining, charge fuel cell
            else:
                p_produced = self.charge_pemfc(abs(p_net))
                n_dcdc_pemfc[t] = p_produced
                p_req = abs(p_net) - p_produced

                # buy remaining electricity from the grid
                e_grid_buy += p_req

                # power that goes through the DC-AC converter,
                # i.e. the load that was covered by the pv array
                n_dcac[t] = (self.load_elec[t] - e_grid_buy)

            # store evolution of storage tank and electricity bought/sold
            self.res['m_h2_array'][t] = ((self.m_h2 - self.m_h2_min) /
                                         (self.m_h2_max - self.m_h2_min))
            self.res['grid_e_buy'][t] = e_grid_buy
            self.res['grid_e_sold'][t] = e_grid_sold

        # define the capacity of the converters and inverter
        self.res['n_dcac'] = max(n_dcac) / 1e3
        self.res['n_dcdc_pemel'] = max(n_dcdc_pemel) / 1e3
        self.res['n_dcdc_pemfc'] = max(n_dcdc_pemfc) / 1e3

        # determine the electrolyzer and fuel cell lifetime
        self.lifetime()

        # determine the self-sufficiency ratio
        self.self_sufficiency_ratio()

        # determine the system cost
        self.cost()

    def lifetime(self):
        """

        The lifetime method determines the lifetime of
        the electrolyzer array and fuel cell array, based on the number of
        operating hours for each component during the evaluated period.

        """
        # lifetime of the electrolyzer array
        if self.res['running_hours_pemel'] == 0.:
            self.res['life_pemel'] = 1e8
        else:
            self.res['life_pemel'] = (self.par['life_pemel'] /
                                      self.res['running_hours_pemel'])

        # lifetime of the fuel cell array
        if self.res['running_hours_pemfc'] == 0.:
            self.res['life_pemfc'] = 1e8
        else:
            self.res['life_pemfc'] = (self.par['life_pemfc'] /
                                      self.res['running_hours_pemfc'])

    def self_sufficiency_ratio(self):
        """

        The self-sufficiency ratio is quantified. The self-sufficiency ratio
        corresponds to the fraction of the electric load that is covered by
        the photovoltaic-hydrogen system.

        """
        self.res['ssr'] = (1. - sum(self.res['grid_e_buy']) /
                           sum(self.load_elec))

    def cost(self):
        """

        Based on the capital recovery factor, the CAPEX,
        OPEX and replacement cost of the system components,
        the levelized cost of electricity is determined. The formula
        for the annualized system cost is adopted from Coppitters et al. [3].

        [3] Coppitters, D., De Paepe, W., & Contino, F. (2020). Robust design
            optimization and stochastic performance analysis of a
            grid-connected photovoltaic system with battery storage and
            hydrogen storage. Energy, 213, 118798.
            https://doi.org/10.1016/j.energy.2020.118798
        """

        # the capital recovery factor
        inv_rate = ((self.par['int_rate'] - self.par['infl_rate']) /
                    (1. + self.par['infl_rate']))
        crf = (((1. + inv_rate)**self.par['life_sys'] - 1.) /
               (inv_rate * (1. + inv_rate)**self.par['life_sys']))**(-1)

        # annual cost of photovoltaic array and DC-DC converter
        pv_cost = self.par['n_pv'] * (crf * self.par['capex_pv'] +
                                      self.par['opex_pv'])
        pv_dcdc_cost = self.res['n_dcdc_pv'] * (self.par['capex_dcdc'] *
                                                (crf + self.par['opex_dcdc']))
        components_cost = pv_cost + pv_dcdc_cost

        # annual cost of electrolyzer array and DC-DC converter
        pemel_cost = self.par['n_pemel'] * (self.par['capex_pemel'] *
                                            (crf + self.par['opex_pemel']))
        pemel_dcdc_cost = (self.res['n_dcdc_pemel'] *
                           (self.par['capex_dcdc'] *
                            (crf + self.par['opex_dcdc'])))
        components_cost += pemel_cost + pemel_dcdc_cost

        # annual cost of fuel cell array and DC-DC converter
        pemfc_cost = self.par['n_pemfc'] * (self.par['capex_pemfc'] * crf +
                                            self.par['opex_pemfc'] *
                                            self.res['running_hours_pemfc'])
        pemfc_dcdc_cost = self.res['n_dcdc_pemfc'] * (self.par['capex_dcdc'] *
                                                      (crf +
                                                       self.par['opex_dcdc']))
        components_cost += pemfc_cost + pemfc_dcdc_cost

        # annual cost of hydrogen storage tank
        tank_cost = self.par['n_tank'] * (self.par['capex_tank'] *
                                          (crf + self.par['opex_tank']))
        components_cost += tank_cost

        # annual cost of DC-AC inverter
        dcac_cost = self.res['n_dcac'] * (self.par['capex_dcac'] *
                                          (crf + self.par['opex_dcac']))
        components_cost += dcac_cost

        # annual replacement cost of the electrolyzer and fuel cell
        arc = crf * sum([(1. + inv_rate)**(-(i + 1.) *
                                           self.res['life_pemel']) *
                         self.par['n_pemel'] *
                         self.par['repl_pemel'] *
                         self.par['capex_pemel'] for i in
                         range(int(self.par['life_sys'] /
                                   self.res['life_pemel']))])
        arc += crf * sum([(1. + inv_rate)**(-(i + 1.) *
                                            self.res['life_pemfc']) *
                          self.par['n_pemfc'] *
                          self.par['repl_pemfc'] *
                          self.par['capex_pemfc'] for i in
                          range(int(self.par['life_sys'] /
                                    self.res['life_pemfc']))])

        grid_e_cost = sum(self.res['grid_e_buy'] * self.elec_profile)
        grid_e_gain = sum(self.res['grid_e_sold'] * self.elec_profile_sale)

        # annual cost
        cost = arc + components_cost + grid_e_cost - grid_e_gain

        # levelized cost of electricity in euro/MWh
        self.res['lcoe'] = cost / (sum(self.load_elec)) * 1e6

    def print_results(self):
        """

        This method prints the levelized cost of electricity,
        the self-sufficiency ratio and the annual energy produced
        by the photovoltaic array.

        """
        print('outputs:')
        print('LCOE:'.ljust(30) + '%.2f euro/MWh' % self.res['lcoe'])
        print('SSR:'.ljust(30) + '%.2f %%' % (self.res['ssr'] * 100.))
        print('PV electricity generated:'.ljust(30) +
              '%.2f MWh' % (sum(self.res['p_pv']) / 1e6))
        print('grid electricity bought:'.ljust(30) +
              '%.2f MWh' % (sum(self.res['grid_e_buy']) / 1e6))
        print('grid electricity sold:'.ljust(30) +
              '%.2f MWh' % (sum(self.res['grid_e_sold']) / 1e6))
        print(
            'lifetime electrolyzer:'.ljust(30) +
            '%.2f year' %
            self.res['life_pemel'])
        print(
            'lifetime fuel cell:'.ljust(30) +
            '%.2f year' %
            self.res['life_pemfc'])

        plt.plot(self.res['m_h2_array'])
        plt.show(block=False)


def polyfit_func(x_in, y_in, threshold=0.9999999):
    """
    The function fits a polynomial to the points of x_in and y_in. The
    polynomial starts with order 1. To evaluate its performance, the R-squared
    performance indicator is quantified. If the value for R-squared does
    not reach the defined threshold, the polynomial order is increased and
    the polynomial is fitted again on the points, until the threshold is
    satisfied. Once satisfied, the function returns the polynomial.

    Parameters
    ----------
    x_in : ndarray
        The x-coordinates for the sample points.
    y_in : ndarray
        The y-coordinates for the sample points.
    threshold : float, optional
        The threshold for the R-squared parameter. The default is 0.99999.

    Returns
    -------
    poly_func : numpy.poly1d
        A one-dimensional polynomial.

    """
    order = 0
    r_squared = 0.
    while r_squared < threshold:
        order += 1

        # the polynomial
        poly_coeff = np.polyfit(x_in, y_in, order)
        poly_func = np.poly1d(poly_coeff)

        # r-squared
        yhat = poly_func(x_in)
        ybar = np.sum(y_in) / len(y_in)
        ssreg = np.sum((yhat - ybar)**2.)
        sstot = np.sum((y_in - ybar)**2.)
        r_squared = ssreg / sstot

    return poly_func



"""
The :py:mod:`genetic_algorithms` module contains a class that performs
the NSGA-II optimization algorithm.
"""

import multiprocessing as mp
import os
import random
from functools import partial
from copy import deepcopy
import numpy as np
from deap import creator, base, tools
import rheia.UQ.pce as uq
import pandas as pd

class NSGA2:
    """

    The NSGAII class includes methods to perform a deterministic and
    robust optimization using NSGA-II optimizer.

    Parameters
    ----------
    run_dict : dict
        The dictionary with user-defined values for
        the characterization of the optimization.
    space_obj : object
        The object with information on the design space
        and stochastic space
    start_from_last_gen : bool
      Boolean that indicates if results exist in the considered result
      directory.
    params : list
        Fixed parameters used during the model evaluation.

    """

    def __init__(self, run_dict, space_obj, start_from_last_gen, params):
        self.run_dict = run_dict
        self.space_obj = space_obj
        self.start_from_last_gen = start_from_last_gen
        self.params = params
        self.my_experiment = None
        self.objective_position = None
        self.opt_res_dir = os.path.join(os.path.abspath(
                                        os.path.join(
                                            os.path.dirname(__file__),
                                            '..')),
                                        'RESULTS',
                                        space_obj.case,
                                        list(run_dict['objectives'].keys())[0],
                                        run_dict['results dir'],
                                        )
    ########################
    # result files methods #
    ########################

    def init_opt(self):
        '''

        Initialization of the results directory and writing of the
        column names in the STATUS file.

        '''

        # check if results directory exists. If not, create one
        if not os.path.exists(self.opt_res_dir):
            os.makedirs(self.opt_res_dir)

        # initialize the STATUS file
        msg = '%8s%8s' % ('ITER', 'EVALS') + '\n' + '%8s%8s' % (0, 0)
        self.write_status(msg)

    def parse_status(self):
        """

        This method extracts the current generation number
        and the number of model evaluations performed.

        Returns
        -------
        ite : int
            The number of generations performed.
        evals : int
            The number of model evaluations performed.

        """

        # the directory of the STATUS file
        stat_dir = os.path.join(self.opt_res_dir, 'STATUS.txt')

        with open(stat_dir, 'rb') as file:

            # read header
            file.readline()

            # read generation number and current number of evaluations
            line = file.readlines()[-1]

            ite, evals = int(line.split()[0]), int(line.split()[1])

        return ite, evals

    def write_status(self, msg):
        """

        A message is appended to the STATUS file.
        This message consists of the generation number
        and computational budget spent.

        Parameters
        ----------
        msg : str
            The message, to be appended to the STATUS file.

        """

        # the directory of the STATUS file
        stat_dir = os.path.join(self.opt_res_dir, 'STATUS.txt')

        # add the message to the STATUS file
        file = open(stat_dir, 'a')
        file.write(msg + '\n')
        file.close()

    def append_points_to_file(self, nests, filename):
        """

        This function is used to append the result (population or fitness)
        to the corresponding file.

        Parameters
        ----------
        nests : list
            The values to be appended to the file.
        filename : str
            The filename where the points
            should be appended.

        """

        # the directory to the file in the results directory
        file_dir = os.path.join(self.opt_res_dir, filename)
                
        dummy = ['-']

        df = pd.DataFrame(nests, columns=None)
        
        with open(file_dir, 'a') as f:
             df.to_csv(f, header=False, index=False, line_terminator='\n')

        df2 = pd.DataFrame(dummy, columns=None)

        with open(file_dir, 'a') as f:
             df2.to_csv(f, header=False, index=False, line_terminator='\n')
        
        '''
        with open(file_dir, 'a') as file:

            for n_in in nests:

                for item in n_in:

                    file.write('%.10f ' % item)

                file.write('\n')

            file.write('- \n')
        '''
        
        
    ##########################################
    # sample creation and evaluation methods #
    ##########################################

    def define_samples_to_eval(self, pop):
        """

        Defines the set of samples considered for evaluation.
        Adds the list of parameters to the initial samples
        drawn form the desing space.

        Parameters
        ----------
        pop : list
            The set of samples (population form, only variables)
            considered for evaluation

        Returns
        -------
        samples_to_eval : list
            The set of samples considered for evaluation
            samples_to_eval = [[p11, p12, ..., p1N, x11, x12, ..., x1M],
                              [p21, p22, ..., p2N, x21, x22, ..., x2M],
                              ...
                              [pk1, pk2, ..., pkN, xk1, xk2, ..., xkM],
                              ]

           where N is the number of parameters
                 M is the number of variables
                 k is the number of samples

        unc_samples_to_eval : list
            The set of samples considered for uncertainty quantification

        """

        # Attach parameters list to the given set of samples
        temp = np.tile(list(self.space_obj.par_dict.values()), (len(pop), 1))

        # check if samples in population have the appropriate length
        for pop_sample in pop:
            if len(pop_sample) != len(self.space_obj.var_dict):
                raise NameError(
                    """A sample in the list of starting design samples
                       does not match the number of design variables.""")

        # store the samples
        temp_samples = np.hstack((temp, np.array(pop))).tolist()
        if 'DET' in self.space_obj.opt_type:
            # return the sample, with deterministic values for the parameters
            # and values for the design variables chosen by the optimizer
            return temp_samples, []

        if 'ROB' in self.space_obj.opt_type:
            # in case of robust optimization, the values for the parameters
            # need to be defined by the PCE algorithm

            # list for the deterministic params and the stochastic params
            samples_to_eval, unc_samples_to_eval = [], []

            # for each design sample, generate the random values for the
            # stochastic params
            for sample in temp_samples:
                my_data = uq.Data(self.run_dict, self.space_obj)

                # acquire information on stochastic parameters
                my_data.read_stoch_parameters(
                    var_values=sample[-len(self.space_obj.var_dict):])

                # array for the number of quantities of interest considered
                self.objective_position = np.zeros(
                    len(self.run_dict['objective names']))
                for index, obj in enumerate(
                        self.run_dict['objective of interest']):

                    # for each quantity of interest, determine its position
                    # in the list
                    self.objective_position[index] = self.run_dict[
                        'objective names'].index(obj)

                # generate a random experiment for the quantity of interest
                self.my_experiment = uq.RandomExperiment(
                    my_data, self.objective_position)

                # create uniform/gaussian distributions and corresponding
                # orthogonal polynomials
                self.my_experiment.create_distributions()

                # determine the number of random samples needed for PCE
                self.my_experiment.n_terms()

                n_samples_unc = self.my_experiment.n_samples

                # no previous results available on the samples for this
                # specific PCE
                self.my_experiment.x_prev = np.array([])
                self.my_experiment.y_prev = np.array([])

                # create a design of experiment for the stochastic params
                self.my_experiment.create_samples(
                    size=self.my_experiment.n_samples)

                # the samples to construct the PCE
                unc_samples = self.my_experiment.x_u

                # produce the final set to evaluate
                temp_unc_samples = np.tile(sample, (n_samples_unc, 1))
                for j, elem in enumerate(self.space_obj.upar_dict.keys()):
                    # find the index of the parameter or variable in the list
                    par_var_list = list(self.space_obj.par_dict.keys())
                    par_var_list += list(self.space_obj.var_dict.keys())
                    unc_idx = par_var_list.index(elem)

                    temp_unc_samples[:, unc_idx] = self.my_experiment.x_u[:, j]

                # convert ndarray to list
                temp_unc_samples = list(temp_unc_samples)

                # add samples to the overall database
                samples_to_eval += temp_unc_samples
                unc_samples_to_eval += list(unc_samples)

            return samples_to_eval, unc_samples_to_eval

    def evaluate_samples(self, samples):
        """

        Evaluation of the set of samples. If the number of jobs
        is larger than 1, parallel processing of the samples
        is considered.

        Parameters
        ----------
        samples : list
            The set of samples considered for evaluation.

        Returns
        -------
        fitness : list
            The calculated fitness for the set of samples.

        """

        eval_dict = []

        # convert the list for each sample into a dictionary
        for sample in samples:
            eval_dict.append(self.space_obj.convert_into_dictionary(sample))

        # linear processing
        if self.run_dict['n jobs'] == 1:
            fitness = []
            for index, sample in enumerate(eval_dict):
                # evaluate the sample dictionary in the evaluate function
                # provide also the index of the sample in the list of samples
                fitness.append(self.run_dict['evaluate']((index, sample),
                                                         params=self.params))

        else:
            # multiprocessing of the sample evaluations
            # provide also the index of the sample in the list of samples
            pool = mp.Pool(processes=self.run_dict['n jobs'])
            fitness = pool.map(partial(self.run_dict['evaluate'],
                                       params=self.params),
                               enumerate(eval_dict))
            pool.close()

        return fitness

    def assign_fitness_to_population(self, pop, fitness, unc_samples):
        """

        Assigns the calulated fitness to the corresponding samples. In case
        of robust optimization, a PCE is constructed first, based on the
        random samples and the corresponding model output. Thereafter, the
        mean and standard deviation are extracted for the quantities of
        interest

        Parameters
        ----------
        pop : list
            Set of samples in population format, considered for evaluation.
        fitness : list
            The calculated fitness.
        unc_samples : list
            The set of samples for uncertainty quantification.

        Returns
        -------
        pop : list
            The population, appended with the fitness values.

        """

        if 'DET' in self.space_obj.opt_type:

            # in case of deterministic optimization, append the deterministic
            # result from the sample evaluations to the population
            for i, sample in enumerate(pop):
                pop[i].fitness.values = fitness[i]

        elif 'ROB' in self.space_obj.opt_type:
            # in case of robust optimization, the mean and standard deviation
            # need to be quantified based on the deterministic result from the
            # sample evaluations

            # Attach parameters list to the sample
            temp = np.tile(list(self.space_obj.par_dict.values()),
                           (len(pop), 1))
            for i in range(len(temp)):

                # store the fitness values temporarily
                temp_fitness = fitness[:self.my_experiment.n_samples]

                # remove these fitness values, as they correspond to a
                # deterministic evaluation of the samples
                del fitness[:self.my_experiment.n_samples]

                # store the random samples in the PCE instance variable
                # for previously generated samples
                self.my_experiment.x_prev = np.array(
                    unc_samples[i * self.my_experiment.n_samples:(i + 1) *
                                self.my_experiment.n_samples])

                # create a design of experiment based on these
                # previously generated samples
                self.my_experiment.create_samples()

                # for each quantity of interest, generate a PCE
                fitness_values = []
                for obj_position in self.objective_position:
                    if obj_position > len(temp_fitness[0]) - 1:
                        raise TypeError(""" The objective "%s" falls out
                                            of the range of predefined
                                            quantities of interest.
                                            Only %i outputs are returned
                                            from the model""" %
                                        (self.run_dict['objective names'][
                                            int(obj_position)],
                                         len(temp_fitness[0])))

                    # add the fitness values to the PCE instance variable
                    # that stores the deterministic results of the sample
                    # evaluations
                    self.my_experiment.y = np.array(temp_fitness)[
                        :, int(obj_position)].reshape(-1, 1)

                    # create a PCE object
                    my_pce = uq.PCE(self.my_experiment)

                    # evaluate the PCE
                    my_pce.run()

                    # Assign the mean and standard deviation as fitness values
                    fitness_values.append(my_pce.moments['mean'])
                    fitness_values.append(
                        np.sqrt(my_pce.moments['variance']))

                pop[i].fitness.values = fitness_values

        return pop

    def read_doe(self, doe_dir):
        """

        Read in the design of experiment.

        Parameters
        ----------
        doe_dir : str
            The directory of the file with the population.

        Returns
        -------
        doe : list
            The Design Of Experiments, i.e. the samples
            to be evaluated in the system model.

        """

        doe = []
        file = open(doe_dir, 'rb')

        # Read doe points
        for line in file:
            doe.append([float(i) for i in line.split()])

            # test if the doe points situate in between the variable bounds
            violate = False
            for index, elem in enumerate(doe[-1]):
                if (elem < self.space_obj.l_b[index] or
                        elem > self.space_obj.u_b[index]):
                    violate = True

            if violate:
                raise TypeError("""Design sample %s violates the
                                   design variable bounds. """ % str(doe[-1]))

        # test if the doe size equals the number of the population size
        # provided in the optimization dictionary
        if len(doe) != self.run_dict['population size']:
            raise TypeError(
                """The number of design samples in the starting population
                   file does not match with the population number provided
                   in the dictionary item "population number". """)

        return doe

    def eval_doe(self):
        """

        Evaluation of the Design Of Experiments (DoE). First, the DoE
        is read from the corresponding population file and stored in
        a list. Then, the design samples are evaluated in the model.
        When a population and fitness values are available from a previous
        optimization run, the method stores the final population and fitness
        values from this previous run. Finally, the population and
        corresponding fitness values are stored in the population and
        fitness file in the results directory, respectively.

        Returns
        -------
        current_pop : list
            The current population.

        """

        # DoE folder path
        path_doe = os.path.join(os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..')),
            'OPT',
            'INPUTS',
            self.space_obj.case,
            '%iD' %
            self.space_obj.n_dim)

        # the DoE file
        file_doe = 'DOE_n%i' % self.run_dict['population size']

        # READ DoE file
        doe = self.read_doe(os.path.join(path_doe, file_doe))

        # evaluate DoE
        n_eval = 0
        current_pop = []
        for i, sol in enumerate(doe):

            # append the instance Individual
            current_pop.append(creator.Individual(sol))

        if not self.start_from_last_gen:
            # in case there is no information from previous results

            # create samples to evaluate
            individuals_to_eval, unc_samples = self.define_samples_to_eval(
                current_pop)

            # evaluate the samples
            fitnesses = self.evaluate_samples(individuals_to_eval)
            n_eval += len(individuals_to_eval)

            # assign fitness to the initial population
            current_pop = self.assign_fitness_to_population(current_pop,
                                                            fitnesses,
                                                            unc_samples)
                                                            
                                                            
            # update the population and fitness files
            self.append_points_to_file(current_pop, 'population.csv')
            
            self.append_points_to_file([x.fitness.values for x in current_pop],
                                       'fitness.csv')

            # update the STATUS file
            self.write_status('%8i%8i' % (1, n_eval))

        else:

            df = pd.read_csv(os.path.join(self.opt_res_dir, 'fitness.csv'))
            df.drop(df.tail(1).index,inplace=True)

            rows = df.tail(self.run_dict['population size']).to_numpy()
            

            output = []
            for line in rows:
                dummy = []
                for l in line:
                    dummy.append(float(l))
                output.append(dummy)

            '''
            # get the population and fitness from the last available generation
            with open(os.path.join(self.opt_res_dir, 'fitness.csv')) as file:

                # Read DOE points
                output = []
                for line in file.readlines()[-len(doe) - 1:-1]:
                    output.append([float(i) for i in line.split()])

            '''
            # add the fitness to the corresponding sample in the population
            for i, pop in enumerate(current_pop):
                pop.fitness.values = output[i]

        return current_pop

    ###################
    # NSGA-II methods #
    ###################

    def nsga2_1iter(self, current_pop):
        """

        Run one iteration of the NSGA-II optimizer.
        Based on the crossover and mutation probabilities, the
        offsprings are created and evaluated. Based on the fitness
        of these offsprings and of the current population,
        a new population is created.

        Parameters
        ----------
        current_pop : list
            The initial population.

        Returns
        -------
        new_pop : list
            The updated population.

        """

        # create offspring, clone of current population
        offspring = [deepcopy(ind) for ind in current_pop]

        # perform crossover
        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < self.run_dict['cx prob']:

                # apply crossover operator
                # in place editing of individuals
                tools.cxSimulatedBinaryBounded(ind1, ind2,
                                               self.run_dict['eta'],
                                               self.space_obj.l_b,
                                               self.space_obj.u_b)

                # set the fitness to an empty tuple of modified individuals
                del ind1.fitness.values
                del ind2.fitness.values

        # perform mutation
        for mutant in offspring:

            if random.random() < self.run_dict['mut prob']:

                # apply mutation operator
                # in place editing of individuals
                tools.mutPolynomialBounded(mutant, self.run_dict['eta'],
                                           self.space_obj.l_b,
                                           self.space_obj.u_b,
                                           self.run_dict['mut prob'])

                # set fitness to empty tuple of modified individuals
                del mutant.fitness.values

        # evaluate the modified individuals
        n_eval = 0
        invalid_indices = []
        invalid_fit_individuals = []
        for i, ind in enumerate(offspring):

            if not ind.fitness.valid:

                invalid_indices.append(i)
                invalid_fit_individuals.append(ind)

        if len(invalid_fit_individuals) > 0:

            # create samples to evaluate
            individuals_to_eval, unc_samples = self.define_samples_to_eval(
                invalid_fit_individuals)

            # evaluate samples
            fitnesses = self.evaluate_samples(individuals_to_eval)
            n_eval += len(individuals_to_eval)

            # assign fitness to the orginal samples list
            individuals_to_assign = self.assign_fitness_to_population(
                invalid_fit_individuals,
                fitnesses,
                unc_samples)

            # construct offspring list
            for i, ind in zip(invalid_indices, individuals_to_assign):

                offspring[i] = deepcopy(ind)

        # select next population using the NSGA-II operator
        
        new_pop = tools.selNSGA2(current_pop + offspring, len(current_pop))

        # update the population and fitness files
        self.append_points_to_file(new_pop, 'population.csv')

        fitness_values = [x.fitness.values for x in new_pop]
        self.append_points_to_file(fitness_values, 'fitness.csv')

        # update the STATUS file
        ite, evals = self.parse_status()
        self.write_status('%8i%8i' % (ite + 1, evals + n_eval))

        return new_pop

    def run_optimizer(self):
        """

        Run an optimization using the NSGA-II algorithm.

        """

        weigth = list(self.space_obj.obj.values())[0]
        creator.create('Fitness', base.Fitness, weights=weigth)
        creator.create('Individual', list, fitness=creator.Fitness)

        if not self.start_from_last_gen:
            self.init_opt()

        # set up the STATUS file
        ite, init_evals = self.parse_status()

        # set up the logbook
        logbook = tools.Logbook()
        logbook.header = 'gen', 'evals', 'min', 'max'
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register('min', np.min, axis=0)
        stats.register('max', np.max, axis=0)

        # evaluate the DoE
        temp_output_pop = self.eval_doe()

        # run the generations
        evals = 0
        while evals < self.run_dict['stop'][1]:

            # perform one NSGA-II iteration
            temp_output_pop = self.nsga2_1iter(temp_output_pop)

            # update the evaluations counter
            ite, current_evals = self.parse_status()
            evals = current_evals - init_evals

            # update the logbook record
            record = stats.compile(temp_output_pop)
            logbook.record(gen=ite, evals=evals, **record)
            print(logbook.stream)


def return_opt_methods():
    """

    Returns the names of the available genetic optimizers.

    Returns
    -------
    name_list : list
        A list with the names of the available genetic optimizers.

    """

    name_list = ['NSGA2']

    return name_list


def return_opt_obj(name):
    """

    Returns the optimizer fuction object

    Parameters
    ----------
    name : str
        Name of the optimizer.

    Returns
    -------
    object
        The optimizer function object.

    """

    switcher = {'NSGA2': NSGA2,
                }

    return switcher.get(name)



"""
The :py:mod:`case_description` module contains a function
to read and store the fixed parameters for model evaluation and a function
to evaluate the system model.
"""


def set_params():
    """
    Set the fixed parameters for each model evaluation.

    """

    params = []

    return params


def evaluate(x_in, params=[]):
    '''
    Evaluation of the system objectives for one given design.

    Parameters
    ----------
    x_in : tuple
        An enumerate object for the input sample.
        The first element of x
        - the index of the input sample in the list of samples -
        can be used for multiprocessing purposes of executable files
        with input and output text files.
        The second element of x - the input sample -
        is a dictionary with the names and values for the model parameters
        and design variables.
    params : list, optional
        List with fixed data, used during model evaluation. The default is [].

    Returns
    -------
    y_1 : float
        model output 1
    y_2 : float
        model output 2
    '''

    # provide your system model and pass x_in as an argument
    y_1 = 1.
    y_2 = 1.

    # return the objectives
    return y_1, y_2



"""
The :py:mod:`case_description` module contains a function
to read and store the fixed parameters for model evaluation and a function
to evaluate the system model.
"""

import os
import rheia.CASES.H2_MOBILITY.h2_mobility as lb


def set_params():
    """
    Set the fixed parameters for each model evaluation.

    """

    path = os.path.dirname(os.path.abspath(__file__))

    filename_climate = os.path.join(os.path.abspath(
                                    os.path.join(path,
                                                 os.pardir)),
                                    'DATA',
                                    'climate',
                                    'climate_Brussels.csv')

    # get the solar irradiance and ambient temperature data
    my_data = lb.ReadData(filename_climate)
    sol_irr, t_amb = my_data.load_climate()

    # store data in params list
    params = [sol_irr, t_amb]

    return params


def evaluate(x_in, params=[]):
    '''
    Evaluation of the system objectives for one given design.

    Parameters
    ----------
    x_in : tuple
        An enumerate object for the input sample.
        The first element of x
        - the index of the input sample in the list of samples -
        can be used for multiprocessing purposes of executable files
        with input and output text files.
        The second element of x - the input sample -
        is a dictionary with the names and values for the model parameters
        and design variables.
    params : list, optional
        List with fixed data, used during model evaluation. The default is [].

    Returns
    -------
    lcom : float
        the levelized cost of driving
    lco2 : float
        the carbon intensity
    grid_e_bought : float
        the annual grid electricity bought
    '''

    arguments = params + [x_in[1]]

    # Evaluation object
    my_evaluation = lb.Evaluation(*arguments)

    # evaluate system model
    succes = my_evaluation.evaluation()
    if succes:

        # when constrained satisfied, get values for objectives
        lcom = my_evaluation.res['lcom']
        ci = my_evaluation.res['ci']
        grid_e_bought = sum(my_evaluation.res['grid_e_buy'])/1e6
    else:

        # when constrained violated, return large value for objectives
        lcom = 1e8
        ci = 1e8
        grid_e_bought = 1e8

    return lcom, ci, grid_e_bought



"""
The :py:mod:`h2_fuel` module contains a class to read the required data and
a class to evaluate the power-to-fuel system.
"""

import os
import pandas as pd
import numpy as np
import pvlib


class ReadData:
    """

    This class enables to read data from the data files.

    Parameters
    ----------
    filename_climate : str
        The directory of the file with information on the
        climate data.

    """

    def __init__(self, filename_climate):
        self.filename_climate = filename_climate
        self.path = os.path.dirname(os.path.abspath(__file__))

    def load_climate(self):
        """

        This method loads the hourly solar irradiance data
        and ambient temperature data,
        situated in the 'sol_irr' and 'T_amb' columns of the
        climate data file.

        Returns
        -------
        sol_irr : ndarray
            The hourly solar irradiance data for a Typical
            Meteorological Year. (8760 elements)
        t_amb : ndarray
            The hourly ambient temperature data for a Typical
            Meteorological Year. (8760 elements)

        """
        data = pd.read_csv(self.filename_climate)
        sol_irr = data['sol_irr'].to_numpy()
        t_amb = data['T_amb'].to_numpy()

        return sol_irr, t_amb

    def load_parameters(self):
        """

        This method loads the deterministic values of the model
        parameters, defined in the design_space file. This is
        useful when the deterministic performance of a specific
        design needs to be evaluated.

        Returns
        -------
        param_dict : dict
            Dictionary with the names of the model parameters
            and the corresponding deterministic values.

        """
        param_dict = {}
        design_space = os.path.join(self.path, 'design_space.csv')

        # read the deterministic values for the parameters in `design_space`
        with open(design_space, 'r') as file:
            for line in file:
                tmp = line.split(",")
                if tmp[1] == 'par':
                    param_dict[tmp[0]] = float(tmp[2])

        return param_dict


class Evaluation:
    """

    This class evaluates the photovoltaic-electrolyzer system.
    For a given design, the solar irradiance, ambient temperature
    and the characterization of the model parameters,
    the levelized cost of hydrogen and the annual hydrogen production
    are quantified.

    Parameters
    ----------
    sol_irr : ndarray
        The hourly solar irradiance for the evaluated year.
    t_amb : ndarray
        The hourly ambient temperature for the evaluated year.
    parameters : dict
        Dictionary with the model parameters and design variables values.

    """

    def __init__(self, sol_irr, t_amb, par):
        self.par = par

        # the solar irradiance and ambient temperature are scaled with the
        # corresponding uncertainty
        self.sol_irr = sol_irr * self.par['u_sol_irr']
        self.t_amb = t_amb + self.par['u_t_amb']

        # the result dictionary
        self.res = {}

        # the system lifetime
        self.par['life_sys'] = 20.

        # the amount of hydrogen produced
        self.res['m_h2'] = 0.

        # initialize the pv power that is consumed by the electrolyzer array
        self.res['p_pv_consumed'] = 0.

        # initialize the operating hours of the electrolyzer array
        self.res['running_hours_pemel'] = 0.

        # the number of PEM electrolyzer cells, corresponding to the
        # nominal capacity of the considered PEM cell and the provided
        # PEM capacity
        self.n_pemel_array = self.par['n_pemel'] / 0.4

        # generate the fitted polynomial on the electrolyzer array
        # power - current relation
        self.polyfit_pemel()

    #############################
    # photovoltaic array module #
    #############################

    def quantify_mpp(self, sol_irr, t_amb, pv_system):
        """

        Quantify the maximum power of the photovoltaic array
        for a given solar irradiance and ambient temperature.

        Parameters
        ----------
        sol_irr : float
            The solar irradiance [W/m2].
        t_amb : float
            The ambient temperature [C].
        pv_system : pandas.core.series.Series
            The pv system characteristics
        Returns
        -------
        pmp : float
            The maximum power.

        """

        # quantify the parameters for the pv system using De Soto method
        pv_inputs = pvlib.pvsystem.calcparams_desoto(sol_irr,
                                                     t_amb,
                                                     pv_system['alpha_sc'],
                                                     pv_system['a_ref'],
                                                     pv_system['I_L_ref'],
                                                     pv_system['I_o_ref'],
                                                     pv_system['R_sh_ref'],
                                                     pv_system['R_s'],
                                                     EgRef=1.121,
                                                     dEgdT=-0.0002677,
                                                     irrad_ref=1000.,
                                                     temp_ref=25.)

        # determine the maximum power for the given pv system
        pmp = pvlib.pvsystem.max_power_point(pv_inputs[0],
                                             pv_inputs[1],
                                             pv_inputs[2],
                                             pv_inputs[3],
                                             pv_inputs[4],
                                             method='newton')['p_mp']

        return pmp

    def photovoltaic(self):
        """

        The hourly photovoltaic power is quantified via the PVlib package.
        Using this package, first the characteristics for a typical
        photovoltaic panel are defined. Based on these characteristics,
        the maximum power point is quantified for each hour, based on the
        corresponding solar irradiance and ambient temperature. Finally, the
        hourly power production is scaled by the considered photovoltaic array
        capacity.

        """

        p_pv = np.zeros(len(self.sol_irr))

        # get the specific photovoltaic panel characteristics
        pv_database = pvlib.pvsystem.retrieve_sam('CECmod')
        pv_system = pv_database.SunPower_SPR_X19_240_BLK

        p_mpp_ref = self.quantify_mpp(1000., 25., pv_system)  # W

        # maximum power point determination for each hour in the timeframe
        for i, irr in enumerate(self.sol_irr):
            if irr > 0.:
                p_mpp = self.quantify_mpp(irr, self.t_amb[i], pv_system)
                p_mpp_array = p_mpp / p_mpp_ref * self.par['n_pv']
                p_pv[i] = min(p_mpp_array, self.par['n_dcdc_pv']) * 1e3  # W
            else:
                p_pv[i] = 0.

        # store the hourly pv power in the result dictionary
        self.res['p_pv'] = p_pv

    #############################
    # electrolyzer array module #
    #############################

    def pemel(self, i_pemel):
        """
        The electrolyzer model, based on the work of Saeed et al. [1]. For a
        given current, the model determines the operating voltage by
        considering the activation, concentration and ohmic overpotentials.
        The model quantifies the operating voltage, power, efficiency and
        hydrogen production.

        [1] Saeed, E. W., & Warkozek, E. G. (2015). Modeling and Analysis of
            Renewable PEM Fuel Cell System. Energy Procedia, 74, 87–101.
            https://doi.org/10.1016/j.egypro.2015.07.527

        Parameters
        ----------
        i_pemel : float
            The electrolyzer input current [A].

        Returns
        -------
        res : dict
            Dictionary with the operating conditions of the electrolyzer for a
            given current. It contains items on the operating voltage, power,
            efficiency and hydrogen mass flow rate.

        """
        par_pemel = {'T': 353.,
                     'a': 1.,
                     'p_o2': 1.,
                     'p_h2': 1.,
                     'p_h2o': 1.,
                     'i_L': 2.,
                     'A': 100.,
                     'i_0': 1e-4,
                     'n': 2.,
                     't_mem': 50e-4,
                     'alpha': 0.3,
                     'R': 8.3143,
                     'F': 96485.,
                     'HHV': 141.7e6,
                     }

        res = {}
        i = i_pemel / par_pemel['A']

        # minimum operating voltage of electrolyzer
        e_0 = (1.48 - 0.85e-3 * (par_pemel['T'] - 298.15) + 4.3085e-5 *
               par_pemel['T'] * np.log(par_pemel['p_h2'] *
                                       np.sqrt(par_pemel['p_o2']) /
                                       par_pemel['p_h2o']))

        # activation overpotential
        v_act = (np.log(i / par_pemel['i_0']) /
                 (par_pemel['alpha'] * par_pemel['n'] * par_pemel['F']) *
                 par_pemel['R'] * par_pemel['T'])

        # ohmic overpotential
        lambda_mem = (0.043 + 17.81 * par_pemel['a'] -
                      39.85 * par_pemel['a']**2. +
                      36. * par_pemel['a']**3.)
        sigma_mem = ((0.005139 * lambda_mem - 0.00326) *
                     np.exp(1268 * (1. / 303. - 1. / par_pemel['T'])))
        v_ohm = i * par_pemel['t_mem'] / sigma_mem

        # the concentration overpotential
        v_con = - (par_pemel['R'] * par_pemel['T'] /
                   (par_pemel['n'] * par_pemel['F']) *
                   np.log(1. - i / par_pemel['i_L']))

        # model outputs
        res['v_pemel'] = (e_0 + v_act + v_ohm + v_con) * self.n_pemel_array
        res['m_pemel'] = self.current_to_mh2(i_pemel) * self.n_pemel_array
        res['p_pemel'] = i_pemel * res['v_pemel']
        res['eff_pemel'] = (res['m_pemel'] * par_pemel['HHV'] /
                          (res['p_pemel'] * 3600.))
        return res

    def current_to_mh2(self, current):
        """
        When current is provided, this function determines the
        corresponding hydrogen mass flow rate per hour.

        Parameters
        ----------
        current : float
            The electrolyzer input current [A].

        Returns
        -------
        m_h2 : float
            The produced hydrogen mass flow rate [kg/h].

        """
        far_cons = 96485.
        m_h2 = current / (2. * far_cons) * 2.02e-3 * 3600.

        return m_h2

    def polyfit_pemel(self):
        """
        The electrolyzer stack is evaluated over a range of input currents.
        Following these evaluations, a polynomial is fitted on the
        power - current relation of the electrolyzer. This polynomial enables
        to rapidly determine the input current when a certain amount of power
        is available. Since this relation is fairly linear, the polynomial
        should reach good agreement with the actual power - current relation,
        while maintaining the level of fidelity of the actual model.

        """

        # evaluate the electrolyzer stack for a set of currents
        i_list = np.arange(start=3, stop=200, step=4)
        p_pemel = np.zeros(len(i_list))
        for index, i in enumerate(i_list):
            res = self.pemel(i)
            p_pemel[index] = res['p_pemel']

        # generate a polynomial fitted on the power - current points
        self.p_to_i_pemel = polyfit_func(p_pemel, i_list)

    def charge_pemel(self, p_pemel):
        """
        For a given power supplied to the electrolyzer, this function
        determines the actual hydrogen produced. First, the method evaluates
        if the power supplied lies within the operating bounds of the
        electrolyzer stack. If the power is situated below the lower limit,
        the electrolyzer does not run. Instead, when the power is situated
        above the upper limit, the electrolyzer operates at nominal conditions.
        At nominal conditions, the current is known and the hydrogen mass flow
        rate is quantified. Otherwise, the input current is determined through
        the fitted polynomial on the power - current relation of the
        electrolyzer stack. As the power is an output from the electrolyzer
        model, considering this polynomial avoids the use of root finding
        methods and is therefore more robust in optimization and uncertainty
        quantification approaches. Finally, when hydrogen is produced, the
        running hours of the electrolyzer stack is increased by 1 and the
        hydrogen mass flow rate is returned.

        Parameters
        ----------
        p_pemel : float
            The power available for electrolysis [W].

        Returns
        -------
        m_h2 : float
            The produced hydrogen mass flow rate [kg/h].

        """
        # the operating bounds
        op_lower_lim = self.par['n_pemel'] * 10.
        op_upper_lim = self.par['n_pemel'] * 1e3

        # check if power is higher than the lowest operating point
        if p_pemel > op_lower_lim:

            # if the power exceeds the upper bound, operate at upper bound
            p_pemel_applied = min(p_pemel, op_upper_lim)

            # current at this applied power
            i_pemel = self.p_to_i_pemel(p_pemel_applied)
            m_h2 = self.current_to_mh2(i_pemel) * self.n_pemel_array
            self.res['p_pv_consumed'] += p_pemel_applied

            # increase the operating hours by 1
            self.res['running_hours_pemel'] += 1.

        # no hydrogen production when the power falls outside the operating
        # bounds
        else:
            m_h2 = 0.

        return m_h2

    #####################
    # evaluation module #
    #####################

    def evaluation(self):
        """

        This is the main method of the Evaluation class.
        In this method, the hourly photovoltaic power is
        quantified first. Then, for each hour, the hydrogen
        is determined. Finally, the electrolyzer lifetime and
        the system cost are determined.

        """

        # get the hourly photovoltaic array power
        self.photovoltaic()

        # evaluate hourly hydrogen production
        for p_pv in self.res['p_pv']:
            self.res['m_h2'] += self.charge_pemel(p_pv)

        # determine the electrolyzer lifetime
        self.lifetime()

        # determine the system cost and levelized cost of hydrogen
        self.cost()

    def lifetime(self):
        """

        The lifetime method determines the lifetime of
        the electrolyzer array, based on the number of
        operating hours during the evaluated year.

        """

        # lifetime of the electrolyzer array
        if self.res['running_hours_pemel'] == 0.:
            self.res['life_pemel'] = 1e8
        else:
            self.res['life_pemel'] = (self.par['life_pemel'] /
                                      self.res['running_hours_pemel'])

    def cost(self):
        """

        Based on the capital recovery factor, the CAPEX,
        OPEX and replacement cost of the system components,
        the levelized cost of hydrogen is determined. The formula
        for the annualized system cost is adopted from Zakeri et al. [2].

        [2] Zakeri, B., & Syri, S. (2015). Electrical energy storage systems:
            A comparative life cycle cost analysis. Renewable and Sustainable
            Energy Reviews, 42, 569–596.
            https://doi.org/10.1016/j.rser.2014.10.011
        """

        # the capital recovery factor
        inv_rate = ((self.par['int_rate'] - self.par['infl_rate']) /
                    (1. + self.par['infl_rate']))
        crf = (((1. + inv_rate)**self.par['life_sys'] - 1.) /
               (inv_rate * (1. + inv_rate)**self.par['life_sys']))**(-1)

        # annual cost of photovoltaic array and DC-DC converter
        pv_cost = self.par['n_pv'] * (crf * self.par['capex_pv'] +
                                      self.par['opex_pv'])
        pv_dcdc_cost = self.par['n_dcdc_pv'] * (self.par['capex_dcdc'] *
                                                (crf + self.par['opex_dcdc']))
        components_cost = pv_cost + pv_dcdc_cost

        # annual cost of electrolyzer array
        pemel_cost = self.par['n_pemel'] * (self.par['capex_pemel'] *
                                            (crf + self.par['opex_pemel']))
        components_cost += pemel_cost

        # annual replacement cost of the electrolyzer array
        arc = crf * sum([(1. + inv_rate)**(-(i + 1.) *
                                           self.res['life_pemel']) *
                         self.par['n_pemel'] *
                         self.par['repl_pemel'] *
                         self.par['capex_pemel'] for
                         i in range(int(self.par['life_sys'] /
                                        self.res['life_pemel']))])

        # the levelized cost of hydrogen
        cost = arc + components_cost
        if self.res['m_h2'] < 1e-5:
            self.res['lcoh'] = 1e8
        else:
            self.res['lcoh'] = cost / self.res['m_h2']

    def print_results(self):
        """

        This method prints the levelized cost of hydrogen,
        the hydrogen production, the annual energy produced
        by the photovoltaic array and the energy consumed by
        the electrolyzer array.

        """

        print('outputs:')
        print('LCOH:'.ljust(30) + '%.2f euro/kg' % self.res['lcoh'])
        print('m_h2:'.ljust(30) + '%.2f kg' % self.res['m_h2'])
        print('PV electricity generated:'.ljust(30) +
              '%.2f MWh' % (sum(self.res['p_pv']) / 1e6))
        print(
            'PV electricity consumed:'.ljust(30) + '%.2f MWh' %
            (self.res['p_pv_consumed'] / 1e6))
        print('self-consumption ratio:'.ljust(30) + '%.2f %%' %
              (1e2 * self.res['p_pv_consumed'] /
               sum(self.res['p_pv'])))
        print('lifetime electrolyzer:'.ljust(30) + '%.2f year' %
              self.res['life_pemel'])


def polyfit_func(x_in, y_in, threshold=0.99999999):
    """
    The function fits a polynomial to the points of x_in and y_in. The
    polynomial starts with order 1. To evaluate its performance, the R-squared
    performance indicator is quantified. If the value for R-squared does
    not reach the defined threshold, the polynomial order is increased and
    the polynomial is fitted again on the points, until the threshold is
    satisfied. Once satisfied, the function returns the polynomial.

    Parameters
    ----------
    x_in : ndarray
        The x-coordinates for the sample points.
    y_in : ndarray
        The y-coordinates for the sample points.
    threshold : float, optional
        The threshold for the R-squared parameter. The default is 0.99999.

    Returns
    -------
    poly_func : numpy.poly1d
        A one-dimensional polynomial.

    """
    order = 0
    r_squared = 0.
    while r_squared < threshold:
        order += 1

        # the polynomial
        poly_coeff = np.polyfit(x_in, y_in, order)
        poly_func = np.poly1d(poly_coeff)

        # r-squared
        yhat = poly_func(x_in)
        ybar = np.sum(y_in) / len(y_in)
        ssreg = np.sum((yhat - ybar)**2.)
        sstot = np.sum((y_in - ybar)**2.)
        r_squared = ssreg / sstot

    return poly_func



'''
Module to test the det_stoch_des_space module.
'''

import pytest
import rheia.OPT.optimization as opt
import rheia.OPT.genetic_algorithms as ga
from deap import creator, base


@pytest.fixture
def run_dict():
    """
    Generate a generic input dictionary for optimization.

    Returns
    -------
    dict_opt : dict
        Input dictionary for optimization.

    """

    dict_opt = {'case': 'H2_FUEL',
                'objectives': {'DET': (-1, 1)},
                'stop': ('BUDGET', 2000),
                'n jobs': 1,
                'population size': 20,
                'results dir': 'run_tutorial',
                }

    return dict_opt


@pytest.fixture
def input_case(run_dict):
    """
    Instantiate a NSGA2 object.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.

    Returns
    -------
    nsga_obj : object
        NSGA2 object.
    space_obj : object
        StochasticDesignSpace object.
    params : list
        List of fixed inputs for evaluation of the model.
    eval_func : function
        Function for evaluation of the model.

    """

    design_space = 'design_space.csv'
    space_obj, eval_func, params = opt.load_case(run_dict, design_space)
    run_dict['evaluate'] = eval_func
    nsga_obj = ga.NSGA2(run_dict, space_obj, True, params)
    return nsga_obj, space_obj, params, eval_func


def test_nsga_obj_run_dict(run_dict, input_case):
    """
    Assert the input dictionary for optimization in NSGA2 object.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    assert input_case[0].run_dict == run_dict


def test_nsga_obj_space_obj(input_case):
    """
    Assert the allocation of the StochasticDesignSpace object in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    assert input_case[0].space_obj == input_case[1]


def test_nsga_obj_start_from_last_gen(input_case):
    """
    Assert the boolean for starting from the last generation in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    assert input_case[0].start_from_last_gen


def test_nsga_obj_params(run_dict, input_case):
    """
    Assert the list with fixed parameters in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    assert input_case[0].params == input_case[2]


def test_parse_status(input_case):
    """
    Assert the status in the STATUS file in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    assert input_case[0].parse_status() == (110, 2000)


def test_define_samples_to_eval(input_case):
    """
    Assert the creation of a population in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    pop = [[2., 5.], [3., 6.], [1., 2.], [7., 2.]]
    assert input_case[0].define_samples_to_eval(pop) == ([[5.0,
                                                           1.0,
                                                           0.0,
                                                           475.0,
                                                           17.5,
                                                           1750.0,
                                                           0.04,
                                                           80000.0,
                                                           0.175,
                                                           150.0,
                                                           0.03,
                                                           0.06,
                                                           0.02,
                                                           2.0,
                                                           5.0],
                                                          [5.0,
                                                           1.0,
                                                           0.0,
                                                           475.0,
                                                           17.5,
                                                           1750.0,
                                                           0.04,
                                                           80000.0,
                                                           0.175,
                                                           150.0,
                                                           0.03,
                                                           0.06,
                                                           0.02,
                                                           3.0,
                                                           6.0],
                                                          [5.0,
                                                           1.0,
                                                           0.0,
                                                           475.0,
                                                           17.5,
                                                           1750.0,
                                                           0.04,
                                                           80000.0,
                                                           0.175,
                                                           150.0,
                                                           0.03,
                                                           0.06,
                                                           0.02,
                                                           1.0,
                                                           2.0],
                                                          [5.0,
                                                           1.0,
                                                           0.0,
                                                           475.0,
                                                           17.5,
                                                           1750.0,
                                                           0.04,
                                                           80000.0,
                                                           0.175,
                                                           150.0,
                                                           0.03,
                                                           0.06,
                                                           0.02,
                                                           7.0,
                                                           2.0]],
                                                         [])


def test_evaluate_samples(input_case):
    """
    Assert the result from evaluation a generic sample in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    samples = [[5.0, 1.0, 0.0, 475.0, 17.5, 1750.0, 0.04,
                80000.0, 0.175, 150.0, 0.03, 0.06, 0.02, 2.0, 5.0]]
    fitness = input_case[0].evaluate_samples(samples)
    assert round(fitness[0][0], 0) == 14
    assert round(fitness[0][1], 0) == 95


def test_assign_fitness_to_population(input_case):
    """
    Assert the assignment of the model output to the fitness in NSGA2 object.

    Parameters
    ----------
    input_case : list
        List that includes the NSGA2 object, StochasticDesignSpace object,
        list with fixed inputs for model evaluation and the function to
        evaluate the model.

    """

    weigth = list(input_case[0].space_obj.obj.values())[0]
    creator.create('Fitness', base.Fitness, weights=weigth)
    creator.create('Individual', list, fitness=creator.Fitness)
    samples = [creator.Individual([5.0,
                                   1.0,
                                   0.0,
                                   475.0,
                                   17.5,
                                   1750.0,
                                   0.04,
                                   80000.0,
                                   0.175,
                                   150.0,
                                   0.03,
                                   0.06,
                                   0.02,
                                   2.0,
                                   5.0])]
    fitness = [[14., 95.]]
    test_pop = input_case[0].assign_fitness_to_population(samples, fitness, [])
    assert test_pop[0].fitness.values == (14., 95.)


def test_return_opt_methods():
    """
    Assert if the NSGA2 method is available in the package.
    """

    methods = ga.return_opt_methods()
    assert methods == ['NSGA2']


def test_return_opt_obj():
    """
    Assert the name for a test object of the NSGA2 class.
    """

    test_object = ga.return_opt_obj('NSGA2')
    assert test_object.__name__ == 'NSGA2'



'''
Module to test the det_stoch_des_space module.
'''

import pytest
import numpy as np
import rheia.UQ.pce as pce
from rheia.CASES.determine_stoch_des_space import load_case


@pytest.fixture
def run_dict():
    """
    Generate a generic input dictionary for uncertainty quantification.

    Returns
    -------
    dict_uq : dict
        Input dictionary for uncertainty quantification.

    """

    dict_uq = {'case': 'H2_FUEL',
               'pol order': 2,
               'objective names': ['lcoh', 'mh2'],
               'objective of interest': 'lcoh',
               'sampling method': 'SOBOL',
               'n jobs': 1,
               'draw pdf cdf': [True, 1e5],
               'results dir': 'opt_design_tutorial'
               }

    return dict_uq


@pytest.fixture
def case(run_dict):
    """
    Instantiate a StochasticDesignSpace object and
    the list of fixed parameters.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for uncertainty quantification.

    Returns
    -------
    space_obj : object
        StochasticDesignSpace object.
    params : list
        List of fixed parameters for evaluation of the model.

    """

    design_space = 'design_space_tutorial_uq.csv'
    space_obj, eval_func, params = load_case(
        run_dict, design_space, uq_bool=True)
    run_dict['evaluate'] = eval_func
    return space_obj, params


@pytest.fixture
def input_case_data(run_dict, case):
    """
    Instantiate a Data object.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for uncertainty quantification.
    case : list
        List that includes the StochasticDesignSpace object and
        list with fixed inputs for model evaluation.

    Returns
    -------
    data_obj : object
        Data object.
    """

    data_obj = pce.Data(run_dict, case[0])
    return data_obj


@pytest.fixture
def input_random_experiment(input_case_data):
    """
    Instantiate a RandomExperiment object.

    Parameters
    ----------
    input_case_data : object
        Data object.

    Returns
    -------
    random_exp_obj : object
        RandomExperiment object.
    """

    input_case_data.read_stoch_parameters()
    input_case_data.create_samples_file()
    random_exp_obj = pce.RandomExperiment(input_case_data, 0)
    return random_exp_obj


@pytest.fixture
def input_pce(input_random_experiment):
    """
    Instantiate a PCE object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.

    Returns
    -------
    pce_obj : object
        PCE object.
    """

    pce_obj = pce.PCE(input_random_experiment)

    return pce_obj


def test_data_inputs(run_dict, input_case_data):
    """
    Assert the input dictionary in the Data object.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for uncertainty quantification.
    input_case_data : object
        Data object.
    """

    assert input_case_data.inputs == run_dict


def test_data_space_obj(input_case_data, case):
    """
    Assert the input dictionary in the Data object.

    Parameters
    ----------
    input_case_data : object
        Data object.
    case : list
        List that includes the StochasticDesignSpace object,
        list with fixed inputs for model evaluation.
    """

    assert input_case_data.space_obj == case[0]


def test_read_stoch_parameters(input_case_data):
    """
    Assert the stochastic data dictionary in the Data object.

    Parameters
    ----------
    input_case_data : object
        Data object.
    """

    input_case_data.read_stoch_parameters()
    input_case_data.stoch_data['l_b'] = [
        round(num, 3) for num in input_case_data.stoch_data['l_b']]
    input_case_data.stoch_data['u_b'] = [
        round(num, 3) for num in input_case_data.stoch_data['u_b']]
    input_case_data.stoch_data['mean'] = [
        round(num, 3) for num in input_case_data.stoch_data['mean']]
    input_case_data.stoch_data['deviation'] = [
        round(num, 3) for num in input_case_data.stoch_data['deviation']]
    assert input_case_data.stoch_data['names'] == [
        'u_sol_irr',
        'u_t_amb',
        'capex_pv',
        'opex_pv',
        'capex_pemel',
        'opex_pemel',
        'life_pemel',
        'repl_pemel',
        'capex_dcdc',
        'opex_dcdc',
        'int_rate',
        'infl_rate']
    assert input_case_data.stoch_data['types'] == [
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform',
        'Uniform']
    assert input_case_data.stoch_data['l_b'] == [
        0.9, -0.4, 350.0, 16.0, 1400.0, 0.03, 60000.0,
        0.15, 100.0, 0.01, 0.04, 0.01]
    assert input_case_data.stoch_data['u_b'] == [
        1.1, 0.4, 600.0, 19.0, 2100.0, 0.05, 100000.0,
        0.2, 200.0, 0.05, 0.08, 0.03]
    assert input_case_data.stoch_data['mean'] == [
        1.0, 0.0, 475.0, 17.5, 1750.0, 0.04, 80000.0,
        0.175, 150.0, 0.03, 0.06, 0.02]
    assert input_case_data.stoch_data['deviation'] == [
        0.1, 0.4, 125.0, 1.5, 350.0, 0.01, 20000.0,
        0.025, 50.0, 0.02, 0.02, 0.01]


def test_random_experiment_my_data(input_random_experiment, input_case_data):
    """
    Assert the Data object in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    input_case_data : object
        Data object.
    """

    assert input_random_experiment.my_data == input_case_data


def test_random_experiment_obj_position(input_random_experiment):
    """
    Assert the objective position in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    """

    assert input_random_experiment.objective_position == 0


def test_random_experiment_dimension(input_random_experiment):
    """
    Assert the stochastic dimension in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    """

    assert input_random_experiment.dimension == 12


def test_random_experiment_n_terms(input_random_experiment):
    """
    Assert the number of terms in the PCE in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    """

    input_random_experiment.n_terms()
    assert input_random_experiment.n_samples == 182


def test_read_prev_samples(input_random_experiment):
    """
    Assert reading the existing samples in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    """

    input_random_experiment.read_previous_samples(False)
    assert len(input_random_experiment.x_prev) == 182
    assert len(input_random_experiment.x_prev[0]) == 12
    assert len(input_random_experiment.y_prev) == 182
    assert input_random_experiment.y_prev[0][0] == 7.598744


def test_create_distr(input_random_experiment):
    """
    Assert the list of polynomial families in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    """

    input_random_experiment.create_distributions()
    assert input_random_experiment.polytypes == ['Legendre'] * 12


def test_create_samples(input_random_experiment):
    """
    Assert the samples size in the RandomExperiment object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    """

    input_random_experiment.read_previous_samples(False)
    input_random_experiment.create_distributions()
    input_random_experiment.create_samples()
    assert input_random_experiment.size == 182
    assert len(input_random_experiment.x_u) == 182
    assert len(input_random_experiment.x_u_scaled) == 182


def test_multindices(input_pce):
    """
    Assert the generic multindices in the PCE object.

    Parameters
    ----------
    input_pce : object
        PCE object.
    """

    assert input_pce.multindices(range(0, 2)) == [
        (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        (1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)]


def test_ols(input_pce):
    """
    Assert the least-square regression method in the PCE object.

    Parameters
    ----------
    input_pce : object
        PCE object.
    """
    a = np.array([[1., -0.96263],
                  [1., -0.94701],
                  [1., 0.0529],
                  [1., 0.5529],
                  ])
    b = np.array([[7.598744],
                  [7.586986],
                  [7.298572],
                  [6.390994]])

    test_res = input_pce.ols(a, b)[0][0]
    assert round(test_res, 3) == 6.996


def test_calc_a(input_random_experiment, input_pce):
    """
    Assert the calculation of the A matrix in the PCE object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    input_pce : object
        PCE object.
    """

    input_random_experiment.read_previous_samples(False)
    input_random_experiment.create_distributions()
    input_random_experiment.create_samples()
    multindices = input_pce.multindices(range(0, 91))
    test_res = input_pce.calc_a(multindices)
    assert round(test_res[0][1], 3) == -0.963
    assert round(test_res[3][12], 3) == -0.533


def test_run(input_random_experiment, input_pce):
    """
    Assert the construction of the PCE in the PCE object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    input_pce : object
        PCE object.
    """

    input_random_experiment.n_terms()
    input_random_experiment.read_previous_samples(False)
    input_random_experiment.create_distributions()
    input_random_experiment.create_samples()
    input_random_experiment.y = input_random_experiment.y_prev
    input_pce.run()
    coeff = input_pce.coefficients
    assert round(coeff[0][0], 3) == 7.730
    assert round(coeff[3][0], 3) == 0.559
    assert round(input_pce.moments['mean'], 3) == 7.730
    assert round(input_pce.moments['variance'], 3) == 0.742


def test_calc_sobol(input_random_experiment, input_pce):
    """
    Assert the calculation of the Sobol' indices in the PCE object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    input_pce : object
        PCE object.
    """

    input_random_experiment.n_terms()
    input_random_experiment.read_previous_samples(False)
    input_random_experiment.create_distributions()
    input_random_experiment.create_samples()
    input_random_experiment.y = input_random_experiment.y_prev
    input_pce.run()
    input_pce.calc_sobol()
    assert round(input_pce.sensitivity['s_i'][0], 3) == 0.059
    assert round(input_pce.sensitivity['s_i'][6], 3) == 0.012
    assert round(input_pce.sensitivity['s_tot_i'][2], 3) == 0.142
    assert round(input_pce.sensitivity['s_tot_i'][9], 3) == 0.002


def test_calc_loo(input_random_experiment, input_pce):
    """
    Assert the calculation of the Leave-One-Out error in the PCE object.

    Parameters
    ----------
    input_random_experiment : object
        RandomExperiment object.
    input_pce : object
        PCE object.
    """

    input_random_experiment.n_terms()
    input_random_experiment.read_previous_samples(False)
    input_random_experiment.create_distributions()
    input_random_experiment.create_samples()
    input_random_experiment.y = input_random_experiment.y_prev
    input_pce.run()
    input_pce.calc_loo()
    assert round(input_pce.loo, 5) == 0.00595



"""
The :py:mod:`h2_mobility` module contains a class to read the required data and
a class to evaluate the power-to-mobility system.
"""

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pvlib


class ReadData:
    """

    This class enables to read data from the data files.

    Parameters
    ----------
    filename_climate : str
        The directory of the file with information on the
        solar irradiance.

    """

    def __init__(self, filename_climate):
        self.filename_climate = filename_climate
        self.path = os.path.dirname(os.path.abspath(__file__))

    def load_climate(self):
        """

        This method loads the hourly solar irradiance data
        and ambient temperature data,
        situated in the 'sol_irr' and 'T_amb' columns of the
        climate data file.

        Returns
        -------
        sol_irr : ndarray
            The hourly solar irradiance data for a Typical
            Meteorological Year.
        t_amb : ndarray
            The hourly ambient temperature data for a Typical
            Meteorological Year.

        """
        data = pd.read_csv(self.filename_climate)
        sol_irr = data['sol_irr'].to_numpy()
        t_amb = data['T_amb'].to_numpy()

        return sol_irr, t_amb

    def load_parameters(self):
        """

        This method loads the deterministic values of the model
        parameters, defined in the design_space file. This is
        useful when the deterministic performance of a specific
        design needs to be evaluated.

        Returns
        -------
        param_dict : dict
            Dictionary with the names of the model parameters
            and the corresponding deterministic values.

        """
        param_dict = {}
        design_space = os.path.join(self.path, 'design_space.csv')

        # read the deterministic values for the parameters in `design_space`
        with open(design_space, 'r') as file:
            for line in file:
                tmp = line.split(",")
                if tmp[1] == 'par':
                    param_dict[tmp[0]] = float(tmp[2])

        return param_dict


class Evaluation:
    """

    This class evaluates the power-to-mobility system.
    For a given design, the solar irradiance, ambient temperature
    and the characterization of the model parameters,
    the levelized cost of driving, carbon intensity and the annual
    grid consumption are quantified.

    Parameters
    ----------
    sol_irr : ndarray
        The hourly solar irradiance for the evaluated year.
    t_amb : ndarray
        The hourly ambient temperature for the evaluated year.
    parameters : dict
        Dictionary with the model parameters and design variables values.

    """

    def __init__(self, sol_irr, t_amb, par):
        self.par = par

        # the solar irradiance and ambient temperature are scaled with the
        # corresponding uncertainty
        self.sol_irr = sol_irr * self.par['u_sol_irr']
        self.t_amb = t_amb + self.par['u_t_amb']

        self.length = len(self.sol_irr)

        # the result dictionary
        self.res = {}

        # the system lifetime
        self.par['life_sys'] = 20.

        # initialize the operating hours of the electrolyzer array
        self.res['running_hours_pemel'] = 0.

        # initialize the storage tank size and its starting status
        self.m_h2_max = self.tank()
        self.m_h2_min = 0.05 * self.m_h2_max
        self.m_h2 = self.m_h2_min

        # instantiate the profiles for grid electricity price
        self.demand_profiles()

        # the number of PEM electrolyzer cells, corresponding to the
        # nominal capacity of the considered PEM cell and the provided
        # PEM capacity
        self.n_pemel_array = self.par['n_pemel'] / 0.4

        # the fitted polynomials on the electrolyzer and compressor
        self.polyfit_pemel()
        self.polyfit_pemel_compr()

    def demand_profiles(self):
        """
        Set the grid electricity price for buying and selling electricity.
        A contract with fixed electricity price is considered, for which the
        price for buying electricity consists of three segments: the energy
        price itself (i.e. 'elec cost'), the profit made on this price by the
        electricity provider (i.e. 'elec_cost_profit') and the fraction of the
        energy price to the final retail price (i.e. 'elec_cost_ratio', e.g.
        when this value equal 0.3, then the energy price corresponds to 30% of
        the final bill, while 70% corresponds to transportation cost,
        taxes,...). The price for selling electricity back to the grid
        corresponds to the energy price.

        In addition, the demand profiles from the hydrogen buses and diesel
        buses is determined, based on the European daily refueling profile [1].

        [1] T. A. Gunawan, I.Williamson, D. Raine, and R. F. Monaghan,
        “Decarbonising city bus networks in ireland with renewable hydrogen,”
        International Journal of Hydrogen Energy, 2020.

        """

        # electricity cost profile [euro/Wh]
        self.elec_profile = np.ones(self.length) * (
            (self.par['elec_cost'] +
             self.par['elec_cost_profit']) /
            self.par['elec_cost_ratio']) / 1e6

        # electricity selling profile [euro/Wh]
        self.elec_profile_sale = np.ones(
            self.length) * self.par['elec_cost'] / 1e6

        self.diesel_profile = np.ones(self.length) * self.par['diesel_cost']

        # number of km driven per day per bus
        self.par['n_km_bus'] = 250.

        # number of buses in the fleet
        self.par['n_bus'] = 40.

        # energy consumed by the diesel buses and hydrogen buses per day
        # [kWh/day]
        energy_h2 = (self.par['cons_h2_bus'] * self.par['n_km_bus'] *
                     self.par['n_h2_bus'])
        energy_diesel = (self.par['cons_diesel_bus'] * self.par['n_km_bus'] *
                         (self.par['n_bus'] - self.par['n_h2_bus']))

        h2_required = energy_h2 / 33.33  # kg
        diesel_required = energy_diesel / 10.  # litre

        # the daily refueling profile of the buses.
        fill_profile = np.array([0.09, 0.015, 0.005, 0.04, 0.04, 0., 0.01,
                                 0.01, 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                                 0.08, 0.08, 0.13, 0.13, 0.13, 0.13, 0.11])

        # daily refueling profile for the hydrogen buses and diesel buses
        day_h2 = np.ones(24) * fill_profile * h2_required
        day_diesel = np.ones(24) * fill_profile * diesel_required

        # annual refueling profiles
        self.load_h2 = list(day_h2) * int(365 * self.length / 8760)
        self.load_diesel = list(day_diesel) * int(365 * self.length / 8760)

        # dispenser capacity such that the hourly hydrogen demand can be
        # complied with
        dispenser_mass_flow_rate = 33.33
        self.par['n_disp'] = max(self.load_h2) / dispenser_mass_flow_rate

    #############################
    # photovoltaic array module #
    #############################

    def quantify_mpp(self, sol_irr, t_amb, pv_system):
        """

        Quantify the maximum power of the photovoltaic array
        for a given solar irradiance and ambient temperature.

        Parameters
        ----------
        sol_irr : float
            The solar irradiance [W/m2].
        t_amb : float
            The ambient temperature [C].
        pv_system : pandas.core.series.Series
            The pv system characteristics
        Returns
        -------
        pmp : float
            The maximum power.

        """

        # quantify the parameters for the pv system using De Soto method
        pv_inputs = pvlib.pvsystem.calcparams_desoto(sol_irr,
                                                     t_amb,
                                                     pv_system['alpha_sc'],
                                                     pv_system['a_ref'],
                                                     pv_system['I_L_ref'],
                                                     pv_system['I_o_ref'],
                                                     pv_system['R_sh_ref'],
                                                     pv_system['R_s'],
                                                     EgRef=1.121,
                                                     dEgdT=-0.0002677,
                                                     irrad_ref=1000.,
                                                     temp_ref=25.)

        # determine the maximum power for the given pv system
        pmp = pvlib.pvsystem.max_power_point(pv_inputs[0],
                                             pv_inputs[1],
                                             pv_inputs[2],
                                             pv_inputs[3],
                                             pv_inputs[4],
                                             method='newton')['p_mp']

        return pmp

    def photovoltaic(self):
        """

        The hourly photovoltaic power is quantified via the PVlib package.
        Using this package, first the characteristics for a typical
        photovoltaic panel are defined. Based on these characteristics,
        the maximum power point is quantified for each hour, based on the
        corresponding solar irradiance and ambient temperature. Finally, the
        hourly power production is scaled by the considered photovoltaic array
        capacity.

        """

        p_pv = np.zeros(len(self.sol_irr))

        # get the specific photovoltaic panel characteristics
        pv_database = pvlib.pvsystem.retrieve_sam('CECmod')
        pv_system = pv_database.SunPower_SPR_X19_240_BLK

        # determine the maximum power point at reference conditions
        p_mpp_ref = self.quantify_mpp(1000., 25., pv_system)  # W

        # maximum power point determination for each hour in the timeframe
        for i, irr in enumerate(self.sol_irr):
            if irr > 0.:
                p_mpp = self.quantify_mpp(irr, self.t_amb[i], pv_system)
                p_pv[i] = p_mpp / p_mpp_ref * self.par['n_pv'] * 1e3  # W
            else:
                p_pv[i] = 0.

        # store the hourly pv power in the result dictionary
        self.res['p_pv'] = p_pv

        # the dc-dc converter capacity in kW
        self.res['n_dcdc_pv'] = max(p_pv) / 1e3

    #############################
    # electrolyzer array module #
    #############################

    def pemel(self, i_pemel):
        """
        The electrolyzer model, based on the work of Saeed et al. [2]. For a
        given current, the model determines the operating voltage by
        considering the activation, concentration and ohmic overpotentials.
        The model quantifies the operating voltage, power, efficiency and
        hydrogen production.

        [2] Saeed, E. W., & Warkozek, E. G. (2015). Modeling and Analysis of
            Renewable PEM Fuel Cell System. Energy Procedia, 74, 87–101.
            https://doi.org/10.1016/j.egypro.2015.07.527

        Parameters
        ----------
        i_pemel : float
            The electrolyzer input current [A].

        Returns
        -------
        res : dict
            Dictionary with the operating conditions of the electrolyzer for a
            given current. It contains items on the operating voltage, power,
            efficiency and hydrogen mass flow rate.

        """
        par_pemel = {'T': 353.,
                     'a': 1.,
                     'p_o2': 1.,
                     'p_h2': 1.,
                     'p_h2o': 1.,
                     'i_L': 2.,
                     'A': 100.,
                     'i_0': 1e-4,
                     'n': 2.,
                     't_mem': 50e-4,
                     'alpha': 0.3,
                     'R': 8.3143,
                     'F': 96485.,
                     'HHV': 141.7e6,
                     }

        res = {}
        i = i_pemel / par_pemel['A']

        # minimum operating voltage of electrolyzer
        e_0 = (1.48 - 0.85e-3 * (par_pemel['T'] - 298.15) + 4.3085e-5 *
               par_pemel['T'] * np.log(par_pemel['p_h2'] *
                                       np.sqrt(par_pemel['p_o2']) /
                                       par_pemel['p_h2o']))

        # activation overpotential
        v_act = (np.log(i / par_pemel['i_0']) /
                 (par_pemel['alpha'] * par_pemel['n'] * par_pemel['F']) *
                 par_pemel['R'] * par_pemel['T'])

        # ohmic overpotential
        lambda_mem = (0.043 + 17.81 * par_pemel['a'] -
                      39.85 * par_pemel['a']**2. +
                      36. * par_pemel['a']**3.)
        sigma_mem = ((0.005139 * lambda_mem - 0.00326) *
                     np.exp(1268 * (1. / 303. - 1. / par_pemel['T'])))
        v_ohm = i * par_pemel['t_mem'] / sigma_mem

        # the concentration overpotential
        v_con = - (par_pemel['R'] * par_pemel['T'] /
                   (par_pemel['n'] * par_pemel['F']) *
                   np.log(1. - i / par_pemel['i_L']))

        # model outputs
        res['v_pemel'] = (e_0 + v_act + v_ohm + v_con) * self.n_pemel_array
        res['m_pemel'] = self.current_to_mh2(i_pemel) * self.n_pemel_array
        res['p_pemel'] = i_pemel * res['v_pemel']
        res['eff_pemel'] = (res['m_pemel'] * par_pemel['HHV'] /
                            (res['p_pemel'] * 3600.))
        return res

    def current_to_mh2(self, current):
        """
        When current is provided, this function determines the
        corresponding hydrogen mass flow rate per hour.

        Parameters
        ----------
        current : float
            The electrolyzer input current [A].

        Returns
        -------
        m_h2 : float
            The produced hydrogen mass flow rate [kg/h].

        """
        far_cons = 96485.
        m_h2 = current / (2. * far_cons) * 2.02e-3 * 3600.

        return m_h2

    def mh2_to_power(self, m_h2):
        """
        When the hydrogen mass flow rate is provided, this function determines
        the corresponding required power per hour.

        Parameters
        ----------
        m_h2 : float
            The produced hydrogen mass flow rate [kg/h].

        Returns
        -------
        power : float
            The required power to produce the hydrogen [W].

        """

        far_cons = 96485.
        current = m_h2 * (2. * far_cons) / (2.02e-3 * 3600. *
                                            self.n_pemel_array)
        power = self.pemel(current)['p_pemel']

        return power

    def polyfit_pemel(self):
        """
        The electrolyzer stack is evaluated over a range of input currents.
        Following these evaluations, a polynomial is fitted on the
        power - current relation of the electrolyzer. This polynomial enables
        to rapidly determine the input current when a certain amount of power
        is available. Since this relation is fairly linear, the polynomial
        should reach good agreement with the actual power - current relation,
        while maintaining the level of fidelity of the actual model.

        """

        # evaluate the electrolyzer stack for a set of currents
        i_list = np.arange(start=3, stop=200, step=4)
        p_pemel = np.zeros(len(i_list))
        for index, i in enumerate(i_list):
            res = self.pemel(i)
            p_pemel[index] = res['p_pemel']

        # generate a polynomial fitted on the power - current points
        self.p_to_i_pemel = polyfit_func(p_pemel, i_list)

    #####################
    # compressor module #
    #####################

    def compressor(self, m_h2):
        """
        The compressor module defined the required compression power to
        compress the hydrogen mass flow rate [3].

        [3] Zhao, L., Brouwer, J., & Samuelsen, S. (2014). Dynamic analysis of
        a self-sustainable renewable hydrogen fueling station. ASME 2014 12th
        International Conference on Fuel Cell Science, Engineering and
        Technology, FUELCELL 2014 Collocated with the ASME 2014 8th
        International Conference on Energy Sustainability.
        https://doi.org/10.1115/FuelCell2014-6330

        Parameters
        ----------
        m_h2 : float
            Hydrogen mass flow rate [kg/h].

        Returns
        -------
        power : float
            The required compression power [W].

        """

        # convert the flow rate into kg/s
        m_h2 *= 1. / 3600.

        par_c = {
            'T_in': 353.,
            'p_in': 20.,
            'p_out': 440.,
            'eta_c': 0.85,
            'R': 4.124,
            'n': 1.609,
        }

        power = (m_h2 *
                 par_c['n'] *
                 par_c['R'] *
                 par_c['T_in'] *
                 ((par_c['p_out'] /
                   par_c['p_in'])**((par_c['n'] -
                                     1.) /
                                    par_c['n']) -
                     1.) *
                 1000. /
                 (par_c['eta_c'] *
                     (par_c['n'] -
                      1.)))

        return power

    def polyfit_pemel_compr(self):
        """
        The power consumption by the electrolyzer stack and compressor are
        evaluated over a range of hydrogen mass flow rates. Following these
        evaluations, a polynomial is fitted on the mass flow rate - power
        relation. This polynomial enables to rapidly determine the input
        mass flow rate when a certain amount of power is available.
        Since this relation is fairly linear, the polynomial should reach good
        agreement with the actual mass flow rate - power relation, while
        maintaining the level of fidelity of the actual model.

        """

        # the electrolyzer array operating limits
        pemel_lower_lim = self.par['n_pemel'] * 10.
        pemel_upper_lim = self.par['n_pemel'] * 1e3

        # the operating current at these limits
        current_ub = self.p_to_i_pemel(pemel_upper_lim)
        current_lb = self.p_to_i_pemel(pemel_lower_lim)

        # the characteristics of the electrolyzer at the limits
        pemel_ub = self.pemel(current_ub)
        pemel_lb = self.pemel(current_lb)

        # the compression power at the hydrogen production limits
        p_compr_ub = self.compressor(pemel_ub['m_pemel'])
        p_compr_lb = self.compressor(pemel_lb['m_pemel'])

        # the compression and electrolyzer power at the limits
        pemel_compr_ub = pemel_ub['p_pemel'] + p_compr_ub
        pemel_compr_lb = pemel_lb['p_pemel'] + p_compr_lb

        # the operating bounds for the compressor and electrolyzer
        self.bounds = {'pemel_lb': pemel_lb,
                       'pemel_ub': pemel_ub,
                       'pemel_compr_lb': pemel_compr_lb,
                       'pemel_compr_ub': pemel_compr_ub
                       }

        # evaluate the electrolyzer and compressor for a set of mass flow rates
        step = pemel_ub['m_pemel'] / 50.
        m_h2_list = np.arange(start=step, stop=pemel_ub['m_pemel'] - step,
                              step=step)
        p_pemel_compr = np.zeros(len(m_h2_list))
        for index, m_h2 in enumerate(m_h2_list):
            p_pemel = self.mh2_to_power(m_h2)
            p_compr = self.compressor(m_h2)
            p_pemel_compr[index] = p_pemel + p_compr

        # generate a polynomial fitted on the power - mass flow rate points
        self.p_to_m_pemel_comp = polyfit_func(p_pemel_compr, m_h2_list)

    ###############
    # tank module #
    ###############

    def tank(self):
        """

        The maximum storage capacity of the hydrogen tank.

        Returns
        -------
        m_max : float
            The hydrogen storage capacity. [kg]

        """

        # conversion from energy (kWh) into mass (kg)
        m_max = self.par['n_tank'] / 33.33

        return m_max

    ##############################
    # management strategy module #
    ##############################

    def p_for_inst_demand(self, m_h2):
        """
        When a hydrogen demand is provided, this method determines the power
        to generate the hydrogen in the electrolyzer array and the power to
        compress the hydrogen in the compressor. The method indicates when the
        electrolyzer array capacity is not sufficient to produce the desired
        hydrogen mass flow rate.

        Parameters
        ----------
        m_h2 : float
            Hydrogen mass flow rate [kg/h].

        Returns
        -------
        p_pemel : float
            The required electrolyzer array power [W].
        p_compr : float
            The required compressor power [W].
        bool
            True when the electrolyzer capacity is insufficient to produce the
            hydrogen.

        """

        # check if the desired hydrogen can be produced by the PEM
        if m_h2 > self.bounds['pemel_ub']['m_pemel']:
            return 0., 0., True

        # determine the power needed from electrolyzer and compressor to
        # deliver the required hydrogen mass flow rate
        p_pemel = self.mh2_to_power(m_h2)
        p_compr = self.compressor(m_h2)
        self.res['running_hours_pemel'] += 1.
        return p_pemel, p_compr, False

    def prod_mh2(self, p_in):
        """
        When there is power available to produce hydrogen, this method
        distributes this power over the electrolyzer array and compressor,
        such that the hydrogen is produced and compressed with this given
        power. If the given power will lead to an excess of hydrogen (i.e
        the storage tank is full before the energy is fully consumed), the
        power supplied to the electrolyzer array and compressor is recalculated
        such that no excess hydrogen is produced.

        Parameters
        ----------
        p_in : float
            The power available to produce hydrogen [W].

        Returns
        -------
        p_pemel : float
            The power consumed by the electrolyzer array [W].
        p_compr : float
            The power consumed by the compressor [W].

        """

        no_run = False

        # produce H2 only if tank is not yet full
        # nothing happens when the desired power is under the lower limit
        if self.m_h2 >= self.m_h2_max or p_in < self.bounds['pemel_compr_lb']:
            p_pemel = 0.
            p_compr = 0.
            no_run = True  # the electrolyzer did not run

        # if the power is higher than the upper limit, work at the upper limit
        elif p_in > self.bounds['pemel_compr_ub']:

            # this is what can be produced at the upper limit for H2
            m_h2 = self.bounds['pemel_ub']['m_pemel']

            # produced hydrogen is added to the tank
            self.m_h2 += m_h2

        else:

            # quantify the hydrogen created with the available power
            m_h2 = self.p_to_m_pemel_comp(p_in)

            # produced hydrogen is added to the tank
            self.m_h2 += m_h2

        # if the new hydrogen capacity exceeds the maximum storage capacity
        if self.m_h2 > self.m_h2_max:

            # the hydrogen that is still allowed in the tank
            m_h2 -= (self.m_h2 - self.m_h2_max)

            # the power for the PEM to generate this hydrogen
            p_rev = self.mh2_to_power(m_h2)

            # check if this power is higher than the lower limit for the PEM
            if p_rev > self.bounds['pemel_lb']['p_pemel']:
                self.m_h2 = self.m_h2_max

            # the space left in the tank is too small
            else:
                self.m_h2 -= (m_h2 + self.m_h2 - self.m_h2_max)
                p_pemel = 0.
                p_compr = 0.
                no_run = True

        # if power is applied to the PEM and compessor
        if not no_run:
            p_compr = self.compressor(m_h2)
            p_pemel = self.mh2_to_power(m_h2)
            self.res['running_hours_pemel'] += 1.

        return p_pemel, p_compr

    def extract_h2_from_tank(self, demand):
        """
        Extract the hydrogen demand from the storage tank. If more hydrogen is
        demanded than available in the tank, extract only the available amount
        of hydrogen.

        Parameters
        ----------
        demand : float
            The hydrogen demand. [kg]

        Returns
        -------
        demand_left : float
            The amount of hydrogen demand that is not covered by the tank. [kg]

        """

        # extract the hydrogen from the tank
        self.m_h2 -= demand

        if self.m_h2 < self.m_h2_min:

            # if the demand was too high, set the tank to its minimum level and
            # define the demand that was not covered by the tank
            demand_left = self.m_h2_min - self.m_h2
            self.m_h2 = self.m_h2_min
        else:

            # if the tank covers the demand, there is no hydrogen demand left
            demand_left = 0.

        return demand_left

    ##########################################
    # model evaluation
    ##########################################

    def evaluation(self):
        """

        This is the main method of the Evaluation class.
        For each hour, the power management strategy is applied.
        The hydrogen demand is extracted from the hydrogen tank, when
        sufficient hydrogen is available in the tank. When the hydrogen in the
        tank does not comply with the hydrogen demand, the power to run the
        electrolyzer array and compressor is calculated to generate and
        compress the remaining hydrogen. To generate this power,
        the photovoltaic power is called upon first. If
        necessary, the remaining power is covered by the grid.
        When excess photovoltaic power is present, this power is used to
        generate and compress hydrogen in the electrolyzer array and
        compressor, respectively. Finally, the lifetime, cost and
        CO2-emission of the system are quantified.

        Returns
        -------
        bool
            True when the capacity of the electrolyzer array is sufficient to
            cover the instantaneous hydrogen demand during the year.

        """

        n_compr = np.zeros(self.length)
        n_dcdc_pem = np.zeros(self.length)
        n_dcac = np.zeros(self.length)

        self.res['m_h2_array'] = np.ones(self.length)
        self.res['grid_e_buy'] = np.ones(self.length)
        self.res['grid_e_sold'] = np.ones(self.length)
        self.res['grid_co2'] = 0.

        # get the hourly photovoltaic array power
        self.photovoltaic()

        for t in range(self.length):
            e_grid_buy = 0.
            e_grid_sold = 0.

            # define if there is any H2 demand left after assessing the tank
            demand_left = self.extract_h2_from_tank(self.load_h2[t])

            if demand_left > 0.:

                # power needed by the PEM and compressor to generate the demand
                p_pemel, p_compr, check = self.p_for_inst_demand(demand_left)
                n_dcdc_pem[t] = p_pemel
                n_compr[t] = p_compr

                # check if PEM capacity can cover the demand
                if check:
                    return False

                # positive if the PV energy can comply with the PEM power
                net_p_1 = self.res['p_pv'][t] - p_pemel

                if net_p_1 > 0.:
                    # if yes, check if remaining PV power can comply with the
                    # compressor
                    net_p_2 = net_p_1 - p_compr

                    if net_p_2 > 0.:
                        # when still excess power available, sell to the grid
                        e_grid_sold = net_p_2
                        n_dcac[t] = e_grid_sold

                    else:
                        # if the compressor demand cannot be
                        # covered, buy remaining demand from the grid
                        e_grid_buy = abs(net_p_2)

                else:
                    # if PEM and compressor cannot be covered by
                    # PV energy, buy required electricity from the grid
                    p_to_buy = abs(net_p_1) + p_compr
                    e_grid_buy = p_to_buy
                    n_dcac[t] = abs(net_p_1)

            else:  # excess PV energy available to generate hydrogen

                if self.res['p_pv'][t] > 0.:
                    # quantify how much of the remaining PV energy can be used
                    # by the PEM and compressor to generate hydrogen
                    p_pemel, p_compr = self.prod_mh2(self.res['p_pv'][t])
                    n_compr[t] = p_compr

                    # the remaining excess PV energy can be sold
                    e_grid_sold = self.res['p_pv'][t] - p_pemel - p_compr
                    n_dcac[t] = e_grid_sold + p_compr
                    n_dcdc_pem[t] = p_pemel

            # store evolution of storage tank, electricity bought/sold
            # and the CO2 amount from buying grid electricity
            self.res['m_h2_array'][t] = ((self.m_h2 - self.m_h2_min) /
                                         (self.m_h2_max - self.m_h2_min))
            self.res['grid_e_buy'][t] = e_grid_buy
            self.res['grid_e_sold'][t] = e_grid_sold
            self.res['grid_co2'] += e_grid_buy * self.par['co2_elec']

        # define the capacity of the converters and compression
        self.res['n_compr'] = max(n_compr) / 1e3
        self.res['n_dcdc_pemel'] = max(n_dcdc_pem) / 1e3
        self.res['n_dcac'] = max(n_dcac) / 1e3

        # the cost of annual diesel consumption
        self.res['diesel_cost'] = self.load_diesel * self.diesel_profile

        # determine the lifetime of the electrolyzer
        self.lifetime()

        # determine the system cost
        self.cost()

        # determine the system annual CO2 emission
        self.lca()

        return True

    def lifetime(self):
        """

        The lifetime method determines the lifetime of
        the electrolyzer array, based on the number of
        operating hours during the evaluated year.

        """

        # lifetime of the electrolyzer array
        if self.res['running_hours_pemel'] == 0.:
            self.res['life_pemel'] = 1e8
        else:
            self.res['life_pemel'] = (self.par['life_pemel'] /
                                      self.res['running_hours_pemel'])

    def cost(self):
        """

        Based on the capital recovery factor, the CAPEX,
        OPEX and replacement cost of the system components,
        the levelized cost of mobility [euro/km] is determined. The formula
        for the annualized system cost is adopted from Coppitters et al. [5].

        [5] Coppitters, D., De Paepe, W., & Contino, F. (2020). Robust design
            optimization and stochastic performance analysis of a
            grid-connected photovoltaic system with battery storage and
            hydrogen storage. Energy, 213, 118798.
            https://doi.org/10.1016/j.energy.2020.118798
        """

        # the capital recovery factor
        inv_rate = ((self.par['int_rate'] - self.par['infl_rate']) /
                    (1. + self.par['infl_rate']))
        crf = (((1. + inv_rate)**self.par['life_sys'] - 1.) /
               (inv_rate * (1. + inv_rate)**self.par['life_sys']))**(-1)

        # annual cost of photovoltaic array and DC-DC converter
        pv_cost = self.par['n_pv'] * (crf * self.par['capex_pv'] +
                                      self.par['opex_pv'])
        pv_dcdc_cost = self.res['n_dcdc_pv'] * (self.par['capex_dcdc'] *
                                                (crf + self.par['opex_dcdc']))
        components_cost = pv_cost + pv_dcdc_cost

        # annual cost of electrolyzer array and DC-DC converter
        pemel_cost = self.par['n_pemel'] * (self.par['capex_pemel'] *
                                            (crf + self.par['opex_pemel']))
        pemel_dcdc_cost = (self.res['n_dcdc_pemel'] *
                           (self.par['capex_dcdc'] *
                            (crf + self.par['opex_dcdc'])))
        components_cost += pemel_cost + pemel_dcdc_cost

        # annual cost of hydrogen storage tank
        tank_cost = self.par['n_tank'] * (self.par['capex_tank'] *
                                          (crf + self.par['opex_tank']))
        components_cost += tank_cost

        # annual cost of compressor
        compressor_cost = (self.par['capex_compr'] *
                           self.res['n_compr'] *
                           (crf + self.par['opex_compr']))
        components_cost += compressor_cost

        # annual cost of dispenser
        dispenser_cost = self.par['n_disp'] * (self.par['capex_disp'] *
                                               (crf + self.par['opex_disp']))
        components_cost += dispenser_cost

        # annual cost of DC-AC inverter
        dcac_cost = self.res['n_dcac'] * (self.par['capex_dcac'] *
                                          (crf + self.par['opex_dcac']))
        components_cost += dcac_cost

        # annual cost of buses
        h2_bus_cost = (self.par['capex_h2_bus'] * crf + self.par['n_km_bus'] *
                       365. * self.par['opex_h2_bus']) * self.par['n_h2_bus']
        diesel_bus_cost = (self.par['capex_diesel_bus'] * crf +
                           self.par['opex_diesel_bus'] * 365. *
                           self.par['n_km_bus']) * (self.par['n_bus'] -
                                                    self.par['n_h2_bus'])
        components_cost += h2_bus_cost + diesel_bus_cost

        # annual replacement cost of the electrolyzer and buses
        arc = crf * sum([(1. + inv_rate)**(-(i + 1.) *
                                           self.res['life_pemel']) *
                         self.par['n_pemel'] *
                         self.par['repl_pemel'] *
                         self.par['capex_pemel'] for
                         i in range(int(self.par['life_sys'] /
                                        self.res['life_pemel']))])
        arc += crf * sum([(1. + inv_rate)**(-(i + 1.) * 10.) *
                          self.par['n_h2_bus'] * self.par['repl_h2_bus']
                          for i in range(int(self.par['life_sys'] / 10.))])
        arc += crf * sum([(1. + inv_rate)**(-(i + 1.) * 10.) *
                          (self.par['n_bus'] - self.par['n_h2_bus']) *
                          self.par['repl_diesel_bus'] for i in
                          range(int(self.par['life_sys'] / 10.))])

        grid_e_cost = sum(self.res['grid_e_buy'] * self.elec_profile)
        grid_e_gain = sum(self.res['grid_e_sold'] * self.elec_profile_sale)

        # total annual cost
        cost = (arc + components_cost + grid_e_cost - grid_e_gain +
                sum(self.res['diesel_cost']))

        # levelized cost of mobility in euro/km
        self.res['lcom'] = cost / (self.par['n_km_bus'] * self.par['n_bus'] *
                                   365.)

    def lca(self):
        """
        The life cycle assessment is performed based on the CO2 emissions from
        constructing the system components and the emissions related to
        consuming grid electricity and diesel. The annual CO2-equivalent
        emissions of the system is divided by the annual distance covered by
        the bus fleet, resulting in the levelized cost of mobility.

        """

        # annual CO2 emission of photovoltaic array production
        pv_co2 = self.par['n_pv'] * self.par['co2_pv']
        pv_dcdc_co2 = self.res['n_dcdc_pv'] * self.par['co2_dcdc']
        comp_co2 = pv_co2 + pv_dcdc_co2

        # annual CO2 emission of electrolyzer array production
        pemel_co2 = (self.par['n_pemel'] * self.par['co2_pemel'] *
                     (1 + int(self.par['life_sys'] / self.res['life_pemel'])))
        pemel_dcdc_co2 = self.res['n_dcdc_pemel'] * self.par['co2_dcdc']
        comp_co2 += pemel_co2 + pemel_dcdc_co2

        # annual CO2 emission of hydrogen storage tank production
        tank_co2 = self.par['n_tank'] * self.par['co2_tank']
        comp_co2 += tank_co2

        # annual CO2 emission of compressor production
        compressor_co2 = self.res['n_compr'] * self.par['co2_compr']
        comp_co2 += compressor_co2

        # annual CO2 emission of dispenser production
        dispenser_co2 = self.par['n_disp'] * self.par['co2_disp']
        comp_co2 += dispenser_co2

        # annual CO2 emission of DC-AC inverter production
        dcac_co2 = self.res['n_dcac'] * self.par['co2_dcac']
        comp_co2 += dcac_co2

        # annual CO2 emission of diesel and hydrogen engine production
        diesel_engine_co2 = (self.par['co2_diesel_engine'] * 200. *  # 200 kW
                             (self.par['n_bus'] - self.par['n_h2_bus']))
        h2_engine_co2 = self.par['co2_fc_engine'] * 200. * self.par['n_h2_bus']
        comp_co2 += ((diesel_engine_co2 + h2_engine_co2) *
                     (1 + int(self.par['life_sys'] / 11.)))  # 10y lifetime

        # annual CO2 emission of diesel consumption
        diesel_co2 = sum(self.load_diesel) * self.par['co2_diesel']

        # annual CO2 emission of system
        co2 = (comp_co2 / self.par['life_sys'] + self.res['grid_co2'] +
               diesel_co2)

        # CO2 emitted per km driven by the fleet
        self.res['ci'] = co2 / (self.par['n_km_bus'] * self.par['n_bus'] *
                                  365. * self.length / 8760.)

    def print_results(self, succes=True):
        """

        This method prints the levelized cost of electricity,
        the self-sufficiency ratio and the annual energy produced
        by the photovoltaic array.

        """
        if not succes:
            print("""Evaluation failed: the electrolyzer array capacity
                     of %f kW was not sufficient to cover the instantaneous
                     hydrogen demand.""" % self.par['n_pemel'])
        else:
            print('outputs:')
            print('LCOE:'.ljust(30) + '%.5f euro/km' % self.res['lcom'])
            print('CI:'.ljust(30) + '%.5f kg co2-eq/km' % self.res['ci'])
            print(
                'PV electricity generated:'.ljust(30) +
                '%.5f MWh' %
                (sum(self.res['p_pv']) / 1e6))
            print('grid energy bought:'.ljust(30) + '%.5f MWh' %
                  (sum(self.res['grid_e_buy']) / 1e6))
            print('grid energy sold:'.ljust(30) + '%.5f MWh' %
                  (sum(self.res['grid_e_sold']) / 1e6))
            print(
                'compressor capacity:'.ljust(30) +
                '%.5f kW' %
                self.res['n_compr'])
            print('life electrolyzer:'.ljust(30) + '%.5f year' %
                  self.res['life_pemel'])

            plt.plot(self.res['m_h2_array'])
            plt.show(block=False)


def polyfit_func(x_in, y_in, threshold=0.99999999):
    """
    The function fits a polynomial to the points of x_in and y_in. The
    polynomial starts with order 1. To evaluate its performance, the R-squared
    performance indicator is quantified. If the value for R-squared does
    not reach the defined threshold, the polynomial order is increased and
    the polynomial is fitted again on the points, until the threshold is
    satisfied. Once satisfied, the function returns the polynomial.

    Parameters
    ----------
    x_in : ndarray
        The x-coordinates for the sample points.
    y_in : ndarray
        The y-coordinates for the sample points.
    threshold : float, optional
        The threshold for the R-squared parameter. The default is 0.99999.

    Returns
    -------
    poly_func : numpy.poly1d
        A one-dimensional polynomial.

    """
    order = 0
    r_squared = 0.
    while r_squared < threshold:
        order += 1

        # the polynomial
        poly_coeff = np.polyfit(x_in, y_in, order)
        poly_func = np.poly1d(poly_coeff)

        # r-squared
        yhat = poly_func(x_in)
        ybar = np.sum(y_in) / len(y_in)
        ssreg = np.sum((yhat - ybar)**2.)
        sstot = np.sum((y_in - ybar)**2.)
        r_squared = ssreg / sstot

    return poly_func
